import rebound
import warnings
import unittest
import os
import math
import sys
from ctypes import c_uint32

class TestSimulation(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.t = 1.246
        self.sim.add(m=1.)
        self.sim.add(m=1e-3, a=1., e=0.01, omega=0.02, M=0.04, inc=0.1)
    
    def tearDown(self):
        self.sim = None

    def test_status(self):
        sys.stdout = open(os.devnull, 'w')
        self.sim.status()
        sys.stdout.close()
        sys.stdout = sys.__stdout__
    
    def test_escape(self):
        self.sim.exit_max_distance = 0.1
        with self.assertRaises(rebound.Escape):
            self.sim.integrate(1.)

    def test_encounter(self):
        self.sim.exit_min_distance = 1.
        with self.assertRaises(rebound.Encounter):
            self.sim.integrate(1.)
    
    def test_removeall(self):
        del self.sim.particles
        self.assertEqual(self.sim.N,0)
    
    def test_remove_too_many(self):
        with warnings.catch_warnings(record=True) as w: 
            warnings.simplefilter("always")
            self.sim.remove(0)
            self.sim.remove(0)
            self.assertEqual(1,len(w))
        with self.assertRaises(RuntimeError):
            self.sim.remove(0)
    
    def test_remove_variational(self):
        v = self.sim.add_variation()
        with self.assertRaises(RuntimeError):
            self.sim.remove(0)
    
    def test_remove(self):
        self.sim.remove(1)
        self.assertEqual(self.sim.N,1)
    
    def test_remove_keepsorted(self):
        self.sim.remove(1,keepSorted=0)
        self.assertEqual(self.sim.N,1)
    
    def test_removehash(self):
        self.sim.add(m=1e-3, a=1., e=0.01, omega=0.02, M=0.04, inc=0.1)
        self.sim.particles[-1].hash = 99
        self.sim.remove(hash=99)
        self.assertEqual(self.sim.N,2)
        with self.assertRaises(RuntimeError):
            self.sim.remove(hash=99)
        with self.assertRaises(RuntimeError):
            self.sim.remove(hash=99)
        with self.assertRaises(RuntimeError):
            self.sim.remove(hash=-99334)
    
    def test_ascii(self):
        a = self.sim.particles_ascii()
        sim = rebound.Simulation()
        sim.add_particles_ascii(a)
        for i in range(self.sim.N):
            self.assertAlmostEqual(self.sim.particles[i].m,sim.particles[i].m,delta=1e-7)
            self.assertAlmostEqual(self.sim.particles[i].x,sim.particles[i].x,delta=1e-7)
            self.assertAlmostEqual(self.sim.particles[i].vy,sim.particles[i].vy,delta=1e-7)
    
    def test_configure_ghostboxes(self):
        self.sim.configure_ghostboxes(1,1,1)
   
    def test_step(self):
        self.sim.step()
        self.assertNotEqual(self.sim.t, 1.246)

    def test_configure_box(self):
        self.assertEqual(self.sim.root_size,-1.)
        self.sim.configure_box(100.,1,1,1)
        self.assertEqual(self.sim.root_size,100.)
    
    def test_calculate_orbits(self):
        orbits = self.sim.calculate_orbits()
        self.assertAlmostEqual(orbits[0].a,1.,delta=1e-15)
        self.assertAlmostEqual(orbits[0].e,0.01,delta=1e-15)
        self.assertAlmostEqual(orbits[0].omega,0.02,delta=1e-12)
        self.assertAlmostEqual(orbits[0].inc,0.1,delta=1e-15)
        orbits = self.sim.calculate_orbits(primary=self.sim.particles[0])
        self.assertAlmostEqual(orbits[0].a,1.,delta=1e-15)
        self.assertAlmostEqual(orbits[0].e,0.01,delta=1e-15)
        self.assertAlmostEqual(orbits[0].omega,0.02,delta=1e-12)
        self.assertAlmostEqual(orbits[0].inc,0.1,delta=1e-15)
        orbits = self.sim.calculate_orbits(primary=self.sim.calculate_com())
        self.assertAlmostEqual(orbits[0].a,1.,delta=1e-2)
        
    def test_com(self):
        self.sim.move_to_com()
        com = self.sim.calculate_com()
        self.assertAlmostEqual(com.x, 0., delta=1e-15)
    
    def test_com_range(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1., x=2.)
        sim.add(m=2., x=5.)
        com = sim.calculate_com(first=1)
        self.assertAlmostEqual(com.x, 4., delta=1e-15)
        com = sim.calculate_com(last=2)
        self.assertAlmostEqual(com.x, 1., delta=1e-15)
        com = sim.calculate_com(first=1,last=2)
        self.assertAlmostEqual(com.x, 2., delta=1e-15)
        com = sim.calculate_com(first=4, last=-3)
        self.assertAlmostEqual(com.x, 0., delta=1e-15)
    
    def test_jacobi_com(self):
        sim = rebound.Simulation()
        sim.add(m=1., x=1.)
        sim.add(m=1., x=3.)
        sim.add(m=2., x=5.)
        com = sim.particles[1].jacobi_com
        self.assertAlmostEqual(com.x, 1., delta=1e-15)
        com = sim.particles[2].jacobi_com
        self.assertAlmostEqual(com.x, 2., delta=1e-15)
        com = sim.particles[0].jacobi_com
        self.assertAlmostEqual(com.x, 0., delta=1e-15)

    def test_init_megno(self):
        self.sim.init_megno()
        self.assertEqual(self.sim.N,4)
        self.assertEqual(self.sim.N_real,2)
        self.assertEqual(self.sim.calculate_megno(),0)
        self.assertTrue(math.isnan(self.sim.calculate_lyapunov()))
        
    def test_calculate_energy(self):
        self.sim.move_to_com()
        energy = self.sim.calculate_energy()
        self.assertAlmostEqual(energy, -0.5e-3, delta=1e-14)

    def test_calculate_angular_momentum(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-3, a=1., inc=0.3, Omega=0.5)
        sim.add(m=1.e-3, a=3., inc=0.2, Omega = -0.8)
        L0 = sim.calculate_angular_momentum()
        sim.integrate(1.)
        Lf = sim.calculate_angular_momentum()
        for i in range(3):
            self.assertAlmostEqual(abs((Lf[i]-L0[i])/L0[i]), 0., delta=1e-15)

    def test_additional_forces(self):
        def af(sim):
            sim.contents.particles[0].hash = 5
            pass
        self.sim.additional_forces = af
        self.sim.integrate(.1)
        self.assertEqual(self.sim.particles[0].hash.value,5)
        with self.assertRaises(AttributeError):
            self.sim.additional_forces

    def test_post_timestep_modifications(self):
        def ptm(sim):
            sim.contents.particles[0].hash = 6
            pass
        self.sim.post_timestep_modifications = ptm
        self.sim.integrate(.1)
        self.assertEqual(self.sim.particles[0].hash.value,6)
        with self.assertRaises(AttributeError):
            self.sim.post_timestep_modifications

    def test_N(self):
        self.assertEqual(self.sim.N, 2)
    
    def test_N_real(self):
        self.assertEqual(self.sim.N_real, 2)
    
    def test_integrator(self):
        self.assertEqual(self.sim.integrator, "ias15")
        self.sim.integrator = "whfast"
        self.assertEqual(self.sim.integrator, "whfast")
        self.sim.integrator = 1
        self.assertEqual(self.sim.integrator, "whfast")
        with self.assertRaises(ValueError):
            self.sim.integrator = "bogusintegrator"

    def test_boundaries(self):
        self.sim.boundary = "open"
        self.assertEqual(self.sim.boundary, "open")
        self.sim.boundary = 8
        self.assertEqual(self.sim.boundary, 8)
        with self.assertRaises(ValueError):
            self.sim.boundary = "bogusboundary"

    
    def test_gravity(self):
        self.sim.gravity = "tree"
        self.assertEqual(self.sim.gravity, "tree")
        self.sim.gravity = 8
        self.assertEqual(self.sim.gravity, 8)
        with self.assertRaises(ValueError):
            self.sim.gravity = "bogusgravity"
    
    def test_collision(self):
        self.sim.collision = "tree"
        self.assertEqual(self.sim.collision, "tree")
        self.sim.collision = 8
        self.assertEqual(self.sim.collision, 8)
        with self.assertRaises(ValueError):
            self.sim.collision = "boguscollision"


    def test_nofile(self):
        with self.assertRaises(RuntimeError):
            sim2 = rebound.Simulation("doesnotexist.bin")


    def test_checkpoint(self):
        self.sim.save("bintest.bin")
        sim2 = rebound.Simulation("bintest.bin")
        self.assertEqual(self.sim.particles[1].x, sim2.particles[1].x)
        self.assertEqual(self.sim.particles[1].vx, sim2.particles[1].vx)
        self.assertEqual(self.sim.t, sim2.t)
        self.assertEqual(self.sim.N, sim2.N)
        self.assertEqual(self.sim.integrator, sim2.integrator)
        os.remove("bintest.bin")
    
    def test_checkpoint_ias15_pointers(self):
        self.sim.integrate(1.)
        self.sim.save("bintest.bin")
        self.sim.integrate(5.)
        sim2 = rebound.Simulation("bintest.bin")
        sim2.integrate(5.)
        self.assertEqual(self.sim.particles[1].x, sim2.particles[1].x)
        self.assertEqual(self.sim.particles[1].vx, sim2.particles[1].vx)
        self.assertEqual(self.sim.t, sim2.t)
        self.assertEqual(self.sim.N, sim2.N)
        self.assertEqual(self.sim.integrator, sim2.integrator)
        os.remove("bintest.bin")
    
class TestSimulationCollisions(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.gravity = "none"
        self.sim.collision = "direct"
        self.sim.collision_resolve = "hardsphere"
        self.sim.integrator = "leapfrog"
        self.sim.G = 0.0
        self.sim.dt = 0.01
    
    def tearDown(self):
        self.sim = None

    def test_coefficient_of_restitution(self):
        self.sim.add(m=1.,x=-1,vx=1.,r=0.5)
        self.sim.add(m=1.,x=1,vx=-1.,r=0.5)
        energy_initial = self.sim.calculate_energy()
        def coef(sim,vrel):
            return 0.5
        self.sim.coefficient_of_restitution = coef
        self.sim.integrate(1.)
        energy_final = self.sim.calculate_energy()
        self.assertAlmostEqual(energy_final, 0.25*energy_initial,delta=1e-15)

    def test_direct(self):
        self.sim.add(m=1.,x=-1,vx=1.,r=0.5)
        self.sim.add(m=1.,x=1,vx=-1.,r=0.5)
        self.sim.integrate(1.)
        self.assertAlmostEqual(self.sim.particles[0].x,-1,delta=1e-15)

    def test_tree(self):
        self.sim.configure_box(10)
        self.sim.collision = "tree"
        self.sim.collision_resolve = "hardsphere"
        self.sim.add(m=1.,x=-1,vx=1.,r=0.5)
        self.sim.add(m=1.,x=1,vx=-1.,r=0.5)
        self.sim.integrate(1.)
        self.assertAlmostEqual(self.sim.particles[0].x,-1,delta=1e-15)

    
    
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Pull data from HORIZONS and format it for use as a REBOUND particle. 

"""
from __future__ import print_function
import telnetlib
import datetime
import re
import sys
import math
import warnings

__all__ = ["getParticle"]

# Default date for orbital elements is the current time when first particle added, if no date is passed.
# Cached at the beginning to ensure that all particles are synchronized.
# If a date is passed, the same date is used for all subsequent particle adds (that don't themselves pass a date).
INITDATE = None

def getParticle(particle=None, m=None, x=None, y=None, z=None, vx=None, vy=None, vz=None, primary=None, a=None, anom=None, e=None, omega=None, inc=None, Omega=None, MEAN=None, date=None, plane="ecliptic"):   
    if plane not in ["ecliptic","frame"]:
        raise AttributeError("Reference plane needs to be either 'ecliptic' or 'frame'. See Horizons for a definition of these coordinate systems.")
    jd = False
    if date is not None:
        if type(date) is datetime.datetime:
            pass
        elif type(date) is str:
            try:
                if date[0:2]=="JD":
                    jd = True
                else:
                    date = datetime.datetime.strptime(date,"%Y-%m-%d %H:%M")
            except:
                raise AttributeError("An error occured while calculating the date. Use either YYYY-MM-DD HH:MM or JDxxxxxxx.xxxxxx")
    # set the cached initialization time if it's not set
    global INITDATE
    if INITDATE is None:
        INITDATE = date if date is not None else datetime.datetime.utcnow()

    if date is None: # if no date passed, used cached value
        date = INITDATE
    if type(date) is datetime.datetime:
        # date is a datetime object
        expect =  (( b'Starting.* :', date.strftime("%Y-%m-%d %H:%M:%S")+'\n' ),
                  ( b'Ending.* :', (date + datetime.timedelta(minutes=1)).strftime("%Y-%m-%d %H:%M")+'\n' ))
    else:
        # Assume date is in JD
        expect =  (( b'Starting.* :', date+'\n' ),
                  ( b'Ending.* :', '\n' ))

    print("Searching NASA Horizons for '%s'... "%(particle),end="")
    sys.stdout.flush()

    t = telnetlib.Telnet()
    t.open('horizons.jpl.nasa.gov', 6775, 20)
    expect = expect + ( ( b'Horizons>', particle+'\n' ),
               ( b'Continue.*:', 'y\n' ),
               ( b'Select.*E.phemeris.*:', 'E\n'),
               ( b'Observe.*:', 'v\n' ),
               ( b'Coordinate center.*:', '@0\n' ),
               ( b'Reference plane.*:', plane+'\n' ),
               ( b'Output interval.*:', '2\n' ),
               ( b'Accept default output \[.*:', 'n\n' ),
               ( b'Output reference frame \[.*:', 'J2000\n' ),
               ( b'Corrections \[.*:', 'NONE\n' ),
               ( b'Output units \[.*:', '1\n' ),
               ( b'Spreadsheet CSV format.*\[.*:', 'NO\n' ),
               ( b'Output delta.*\[.*:', 'NO\n' ),
               ( b'Select output table type.*\[.*:', '3\n' ),
               ( b'Label cartesian output.*\[.*:', 'NO\n' ),
               ( b'Scroll . Page: .*%', ' '),
               ( b'Select\.\.\. .A.gain.* :', 'X\n' ),
               ( b'Select \.\.\. .F.tp.*:', 'selectID' )
    )
    p = Particle()
    startdata = 0
    message = ""
    idn = None
    while True:
        try:
            answer = t.expect(list(i[0] for i in expect), 8)
        except EOFError:
            break
        a = answer[2].decode()
        if "$$SOE" in a:
            lines = a.split("\n")
            for line in lines:
                if line.strip() == "$$EOE":
                    break
                if startdata == 2:
                    pos = [float(i) for i in line.split()]
                    p.x = pos[0]
                    p.y = pos[1]
                    p.z = pos[2]
                if startdata == 3:
                    vel = [float(i) for i in line.split()]
                    p.vx = vel[0]
                    p.vy = vel[1]
                    p.vz = vel[2]
                if startdata > 0:
                    startdata += 1
                if "Target body name:" in line:
                    print("Found: %s." % line[18:50].strip())
                    try:
                        idn = re.search(r"\(([0-9]+)\)", line).group(1)
                    except:
                        pass
                if line.strip() == "$$SOE":
                    startdata = 1
        message += a.replace(chr(27)+"[H","")
        if answer[0] != -1:
            if expect[answer[0]][1] == "selectID":
                try:
                    idn = a.split("ID#")[1].split("\n")[2].split()[0]
                except:
                    try:
                        idn = a.split("Record #")[1].split("\n")[2].split()[0]
                    except:
                        raise Exception("Error while trying to find object.")
                t.write((idn+"\n").encode())
            else:
                t.write(expect[answer[0]][1].encode())
        else:
            pass
    if startdata == 0:
        print(message)
        raise SyntaxError("Object not found. See above output from HORIZONS. Please try different identifier or look up JPL Body Number.")
    if m is not None:
        p.m = m
    elif idn is not None:
        try:
            p.m = float(re.search(r"BODY%d\_GM .* \( *([\.E\+\-0-9]+ *)\)"%int(idn), HORIZONS_MASS_DATA).group(1))
            p.m /= Gkmkgs # divide by G (horizons masses give GM)
        except:
            warnings.warn("Warning: Mass cannot be retrieved from NASA HORIZONS. Set to 0.", RuntimeWarning)
            p.m = 0
    else:
        warnings.warn("Warning: Mass cannot be retrieved from NASA HORIZONS. Set to 0.", RuntimeWarning)
        p.m = 0
    return p


# There is currently no way to get mass data from HORIZONS.
# The following data was provided by Jon Giorgini (10 May 2015)
# Last updated: May 31 2018. 
# Source: ftp://ssd.jpl.nasa.gov/pub/xfr/gm_Horizons.pck
# Units: km^3/s^2

Gkmkgs = 6.67408e-20 # units of km^3/kg/s^2

HORIZONS_MASS_DATA = """
    BODY1_GM       = ( 2.2031780000000021E+04 )
    BODY2_GM       = ( 3.2485859200000006E+05 )
    BODY3_GM       = ( 4.0350323550225981E+05 )
    BODY4_GM       = ( 4.2828375214000022E+04 )
    BODY5_GM       = ( 1.2671276480000021E+08 )
    BODY6_GM       = ( 3.7940585200000003E+07 )
    BODY7_GM       = ( 5.7945486000000080E+06 )
    BODY8_GM       = ( 6.8365271005800236E+06 )
    BODY9_GM       = ( 9.7700000000000068E+02 )
    BODY10_GM      = ( 1.3271244004193938E+11 )

    BODY199_GM     = ( 2.2031780000000021E+04 )
    BODY299_GM     = ( 3.2485859200000006E+05 )
    BODY399_GM     = ( 3.9860043543609598E+05 )
    BODY499_GM     = ( 4.282837362069909E+04  )
    BODY599_GM     = ( 1.266865349115908E+08  )
    BODY699_GM     = ( 3.793120723493890E+07  )
    BODY799_GM     = ( 5.793951322279009E+06  )
    BODY899_GM     = ( 6.835099502439672E+06  )
    BODY999_GM     = ( 8.693390780381926E+02  )
 
    BODY301_GM     = ( 4.9028000661637961E+03 )

    BODY401_GM     = ( 7.087546066894452E-04 )
    BODY402_GM     = ( 9.615569648120313E-05 )

    BODY501_GM     = ( 5.959924010272514E+03 )
    BODY502_GM     = ( 3.202739815114734E+03 )
    BODY503_GM     = ( 9.887819980080976E+03 )
    BODY504_GM     = ( 7.179304867611079E+03 )
    BODY505_GM     = ( 1.487604677404272E-01 )
    BODY506_GM     = ( 1.380080696078966E-01 )
 
    BODY601_GM     = ( 2.503458199931431E+00 )
    BODY602_GM     = ( 7.211185066509890E+00 )
    BODY603_GM     = ( 4.120856508658532E+01 )
    BODY604_GM     = ( 7.311574218947423E+01 )
    BODY605_GM     = ( 1.539419035933117E+02 )
    BODY606_GM     = ( 8.978137030983542E+03 )
    BODY607_GM     = ( 3.712085754472412E-01 )
    BODY608_GM     = ( 1.205095752388872E+02 )
    BODY609_GM     = ( 5.532371285376407E-01 )
    BODY610_GM     = ( 1.265765099012197E-01 )
    BODY611_GM     = ( 3.512333288208074E-02 )
    BODY612_GM     = ( 3.424829447502984E-04 )
    BODY615_GM     = ( 3.718871247516475E-04 )
    BODY616_GM     = ( 1.075208001007610E-02 )
    BODY617_GM     = ( 9.290325122028795E-03 )

    BODY701_GM     = ( 8.346344431770477E+01 )
    BODY702_GM     = ( 8.509338094489388E+01 )
    BODY703_GM     = ( 2.269437003741248E+02 )
    BODY704_GM     = ( 2.053234302535623E+02 )
    BODY705_GM     = ( 4.319516899232100E+00 )

    BODY801_GM     = ( 1.427598140725034E+03 )
  
    BODY901_GM     = ( 1.062509269522026E+02 )
    BODY902_GM     = ( 2.150552267969335E-03 )
    BODY903_GM     = ( 3.663917107480563E-03 )
    BODY904_GM     = ( 4.540734312735987E-04 )
    BODY905_GM     = ( 2.000000000000000E-20 )

    BODY2000001_GM = ( 6.2809393000000000E+01 )
    BODY2000002_GM = ( 1.3923011000000001E+01 )
    BODY2000003_GM = ( 1.6224149999999999E+00 )
    BODY2000004_GM = ( 1.7288008999999999E+01 )
    BODY2000010_GM = ( 5.5423920000000004E+00 )
    BODY2000015_GM = ( 2.0981550000000002E+00 )
    BODY2000016_GM = ( 1.5300480000000001E+00 )
    BODY2000031_GM = ( 2.8448720000000001E+00 )
    BODY2000048_GM = ( 1.1351590000000000E+00 )
    BODY2000052_GM = ( 1.1108039999999999E+00 )
    BODY2000065_GM = ( 1.4264810000000001E+00 )
    BODY2000087_GM = ( 9.8635300000000004E-01 )
    BODY2000088_GM = ( 1.1557990000000000E+00 )
    BODY2000433_GM = ( 4.463E-4 )
    BODY2000451_GM = ( 1.0295259999999999E+00 )
    BODY2000511_GM = ( 2.3312860000000000E+00 )
    BODY2000704_GM = ( 2.3573170000000001E+00 )
"""


# Import at the end to avoid circular dependence
from .particle import *



import rebound
import unittest
import math
import rebound.data
    
whfastsettings1 = [ # corrector, corrector2, kernel, relative error
        [0, 0,"default",2e-7], 
        [3, 0,"default",1e-9], 
        [5, 0,"default",1e-9], 
        [7, 0,"default",1e-9], 
        [11,0,"default",1e-9], 
        [17,0,"default",1e-9], 
        [0, 0,"modifiedkick",2e-7], 
        [3, 0,"modifiedkick",1e-9], 
        [5, 0,"modifiedkick",1e-11], 
        [7, 0,"modifiedkick",1e-12], 
        [11,0,"modifiedkick",1e-12], 
        [17,0,"modifiedkick",1e-12], 
        [17,0,"composition",1e-12], 
        [17,0,"lazy",1e-12], 
        [11,1,"modifiedkick",1e-12], 
        ]

whfastsettings2 = [ # corrector, corrector2, kernel, relative error
        [0, 0,"default",2e-7], 
        [3, 0,"default",1e-9], 
        [3, 0,"modifiedkick",1e-9], 
        [3, 0,"composition",1e-12], 
        [3, 0,"lazy",1e-12], 
        [3, 0,"default",1e-9], 
        [3, 0,"modifiedkick",1e-9], 
        [3, 0,"composition",1e-12], 
        [3, 0,"lazy",1e-12], 
        [3, 1,"lazy",1e-12], 
        ]

class TestIntegratorWHFastAdvanced(unittest.TestCase):
    def test_whfastsettings1(self):
        for s in whfastsettings1:
            test_name = "test_energy_WHFastAdvanced_c_%02d_c2_%d_kernel_%s" % (s[0], s[1], s[2])
            self.energy(s)
            test_name = "test_energy_notcom_WHFastAdvanced_c_%02d_c2_%d_kernel_%s" % (s[0], s[1], s[2])
            self.energy_notcom(s)
            test_name = "test_compias_WHFastAdvanced_c_%02d_c2_%d_kernel_%s" % (s[0], s[1], s[2])
            self.compias(s)
    def test_whfastsettings2(self):
        for s in whfastsettings2:
            test_name = "test_backandforth_WHFastAdvanced_c_%02d_c2_%d_kernel_%s" % (s[0], s[1], s[2])
            self.backandforth(s)
    def test_whfastrestart(self):
        for s in whfastsettings2:
            test_name = "test_restart_WHFastAdvanced_c_%02d_c2_%d_kernel_%s" % (s[0], s[1], s[2])
            self.restart(s)

    def energy(self, s):
        corrector, corrector2, kernel, maxerror = s
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        sim.integrator = "whfast"
        sim.ri_whfast.corrector = corrector 
        sim.ri_whfast.corrector2 = corrector2
        sim.ri_whfast.kernel = kernel
        sim.ri_whfast.safe_mode = False
        sim.dt = 0.0123235235*sim.particles[1].P  
        e0 = sim.calculate_energy()
        sim.integrate(1000.*2.*3.1415)
        e1 = sim.calculate_energy()
        self.assertLess(math.fabs((e0-e1)/e1),maxerror)

    def energy_notcom(self, s):
        corrector, corrector2, kernel, maxerror = s
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        for p in sim.particles:
            p.vx += 1.
        com = sim.calculate_com()
        sim.integrator = "whfast"
        sim.ri_whfast.corrector = corrector 
        sim.ri_whfast.corrector2 = corrector2
        sim.ri_whfast.kernel = kernel
        sim.dt = 0.0123235235*sim.particles[1].P  
        e0 = sim.calculate_energy()
        sim.integrate(1000.*2.*3.1415)
        e1 = sim.calculate_energy()
        self.assertLess(math.fabs((e0-e1)/e1),maxerror)
        com1 = sim.calculate_com()
        self.assertLess(math.fabs((com.x+com.vx*sim.t-com1.x)/(com1.x+com1.y)),1e-12)
        self.assertLess(math.fabs((com.y+com.vy*sim.t-com1.y)/(com1.x+com1.y)),1e-12)

    def compias(self, s):
        corrector, corrector2, kernel, maxerror = s
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        for p in sim.particles:
            p.vx += 1050. # move out of com to make it harder
        sim.integrator = "whfast"
        sim.ri_whfast.corrector = corrector 
        sim.ri_whfast.corrector2 = corrector2
        sim.ri_whfast.kernel = kernel
        sim.dt = 0.0123235235*sim.particles[1].P  
        sim.integrate(13.21415,exact_finish_time=False)
        
        simi = rebound.Simulation()
        simi.integrator = "ias15"
        rebound.data.add_outer_solar_system(simi)
        for p in simi.particles:
            p.vx += 1050.
        simi.integrate(sim.t,exact_finish_time=True)
        
        for i in range(sim.N):
            if corrector:
                self.assertLess(math.fabs(simi.particles[i].x-sim.particles[i].x),1e-8)
            else: # less accurate
               self.assertLess(math.fabs(simi.particles[i].x-sim.particles[i].x),2e-7)

    def backandforth(self, s):
        corrector, corrector2, kernel, maxerror = s
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        for p in sim.particles:
            p.vx += 1.
        sim0=sim.copy()
        sim.integrator = "whfast"
        sim.ri_whfast.corrector = corrector 
        sim.ri_whfast.corrector2 = corrector2
        sim.ri_whfast.kernel = kernel
        sim.dt = 0.0123235235*sim.particles[1].P  
        steps = 10
        for i in range(steps):
            sim.step()
        sim.dt *= -1
        for i in range(steps):
            sim.step()
        for i in range(sim.N):
            if corrector:
                self.assertLess(math.fabs(sim0.particles[i].x-sim.particles[i].x),2e-11)
    
    def restart(self, s):
        corrector, corrector2, kernel, maxerror = s
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        sim.integrator = "whfast"
        sim.ri_whfast.corrector = corrector 
        sim.ri_whfast.corrector2 = corrector2
        sim.ri_whfast.kernel = kernel
        sim.step()
        sim2 = sim.copy()
        sim.step()
        sim2.step()
        self.assertEqual(sim,sim2)

if __name__ == "__main__":
    unittest.main()



from ctypes import Structure, c_double, c_int, byref, memmove, sizeof, c_uint32, c_uint, c_ulong
from . import clibrebound
import math
import ctypes.util
import rebound
import sys
import random

__all__ = ["Particle"]

def notNone(a):
    """
    Returns True if array a contains at least one element that is not None. Returns False otherwise.
    """
    return a.count(None) != len(a)

class Particle(Structure):
    """
    The main REBOUND particle data structure. 
    This is an abstraction of the reb_particle structure in C.
    The Particle fields are set at the end of simulation.py to avoid circular references.
    
    Attributes
    ----------
    x, y, z     : float       
        Particle positions
    vx, vy, vz  : float       
        Particle velocities
    ax, ay, az  : float       
        Particle accelerations
    m           : float       
        Particle mass
    r           : float       
        Particle radius
    lastcollision : float       
        Last time the particle had a physical collision (if checking for collisions)
    c           : c_void_p (C void pointer) 
        Pointer to the cell the particle is currently in (if using tree code)
    hash          : c_uint32         
        Particle hash (permanent identifier for the particle)
    ap          : c_void_p (C void pointer)
        Pointer to additional parameters one might want to add to particles
    _sim        : POINTER(rebound.Simulation)
        Internal pointer to the parent simulation (used in C version of REBOUND)
    a, e, inc, Omega, omega, f	: float
	    (Kepler Elements) Semi-major axis, eccentricity, inclination, longitude of the ascending node, argument of periapsis, and true anomaly respectively. The Keplerian Elements are in Jacobi coordinates (with mu = G*Minc, where Minc is the total mass from index 0 to the particle's index, inclusive).
    """
    def __str__(self):
        """ 
        Returns a string with the position and velocity of the particle.
        """
        return "<rebound.Particle object, m=%s x=%s y=%s z=%s vx=%s vy=%s vz=%s>"%(self.m,self.x,self.y,self.z,self.vx,self.vy,self.vz)
   
    __repr__ = __str__

    def __init__(self, simulation=None, particle=None, m=None, x=None, y=None, z=None, vx=None, vy=None, vz=None, primary=None, a=None, P=None, e=None, inc=None, Omega=None, omega=None, pomega=None, f=None, M=None, l=None, theta=None, T=None, r=None, date=None, variation=None, variation2=None, h=None, k=None, ix=None, iy=None, hash=0, jacobi_masses=False):
        """
        Initializes a Particle structure. Rather than explicitly creating 
        a Particle structure, users may use the ``add()`` member function 
        of a Simulation instance, which will both create a Particle and 
        then add it to the simulation with one function call.

        This function accepts either cartesian positions and velocities, 
        classical orbital elements together with the reference Particle 
        (the primary), as well as orbital parameters defined by Pal (2009).

        For convenience, optional keywords that are not passed default 
        to zero (mass, cartesian and orbital elements). 

        Whenever initializing a particle from orbital elements, one must 
        specify either the semimajor axis or the period of the orbit.
        
        For classical orbital paramerers, one can specify the longitude 
        of the ascending node by passing Omega, to specify the pericenter 
        one can pass either omega or pomega (not both), and for the 
        longitude/anomaly one can pass one of f, M, l or theta.  
        See ipython_examples/OrbitalElements.ipynb for examples.  
        See also Murray & Dermott Solar System Dynamics for formal 
        definitions of angles in orbital mechanics.

        All angles should be specified in radians.
        

        Parameters
        ----------
        simulation  : Simulation  
            Simulation instance associated with this particle (Required if passing orbital elements or setting up a variation).
        particle    : Particle, optional    
            If a particle is passed, a copy of that particle is returned.
            If a variational particle is initialized, then ``particle`` is 
            original particle that will be varied. 
        m           : float       
            Mass        (Default: 0)
        x, y, z     : float       
            Positions in Cartesian coordinates  (Default: 0)
        vx, vy, vz  : float       
            Velocities in Cartesian coordinates (Default: 0)
        primary     : Particle    
            Primary body for converting orbital elements to cartesian (Default: center of mass of the particles in the passed simulation, i.e., this will yield Jacobi coordinates as one progressively adds particles) 
        a           : float       
            Semimajor axis (a or P required if passing orbital elements)
        P           : float
            Orbital period (a or P required if passing orbital elements)
        e           : float       
            Eccentricity                (Default: 0)
        inc         : float       
            Inclination                 (Default: 0)
        Omega       : float       
            Longitude of ascending node (Default: 0)
        omega       : float       
            Argument of pericenter      (Default: 0)
        pomega      : float       
            Longitude of pericenter     (Default: 0)
        f           : float       
            True anomaly                (Default: 0)
        M           : float       
            Mean anomaly                (Default: 0)
        l           : float       
            Mean longitude              (Default: 0)
        theta       : float       
            True longitude              (Default: 0)
        T           : float 
            Time of pericenter passage  
        h           : float       
            h variable, see Pal (2009) for a definition  (Default: 0)
        k           : float       
            k variable, see Pal (2009) for a definition  (Default: 0)
        ix          : float       
            ix variable, see Pal (2009) for a definition  (Default: 0)
        iy           : float       
            iy variable, see Pal (2009) for a definition  (Default: 0)
        r           : float       
            Particle radius (only used for collisional simulations)
        date        : string      
            For consistency with adding particles through horizons.  Not used here.
        variation   : string            (Default: None)
            Set this string to the name of an orbital parameter to initialize the particle as a variational particle.
            Can be one of the following: m, a, e, inc, omega, Omega, f, k, h, lambda, ix, iy.
        variation2  : string            (Default: None)
            Set this string to the name of a second orbital parameter to initialize the particle as a second order variational particle. Only used for second order variational equations. 
            Can be one of the following: m, a, e, inc, omega, Omega, f, k, h, lambda, ix, iy.
        hash        : c_uint32  
            Unsigned integer identifier for particle.  Can pass an integer directly, or a string that will be converted to a hash. User is responsible for assigning unique hashes.
        jacobi_masses: bool
            Whether to use jacobi primary mass in orbit initialization. Particle mass will still be set to physical value (Default: False)
        Examples
        --------

        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> p1 = rebound.Particle(simulation=sim, m=0.001, a=0.5, e=0.01)
        >>> p2 = rebound.Particle(simulation=sim, m=0.0, x=1., vy=1.)
        >>> p3 = rebound.Particle(simulation=sim, m=0.001, a=1.5, h=0.1, k=0.2, l=0.1)
        >>> p4 = rebound.Particle(simulation=sim, m=0.001, a=1.5, omega="uniform")  # omega will be a random number between 0 and 2pi

        """        

        if Omega == "uniform":
            Omega = random.vonmisesvariate(0.,0.) 
        if omega == "uniform":
            omega = random.vonmisesvariate(0.,0.) 
        if pomega == "uniform":
            pomega = random.vonmisesvariate(0.,0.) 
        if f == "uniform":
            f = random.vonmisesvariate(0.,0.) 
        if M == "uniform":
            M = random.vonmisesvariate(0.,0.) 
        if l == "uniform":
            l = random.vonmisesvariate(0.,0.) 
        if theta == "uniform":
            theta = random.vonmisesvariate(0.,0.) 
        if inc == "uniform":
            inc = random.vonmisesvariate(0.,0.) 

        self.hash = hash # set via the property, which checks for type

        if variation:
            if primary is None:
                primary = simulation.particles[0]
            # Find particle to differenciate
            lc = locals().copy()
            del lc["self"]
            del lc["variation"]
            del lc["variation2"]
            if particle is None:
                particle = Particle(**lc)
            # First or second order?
            if variation and variation2:
                variation_order = 2
            else:
                variation_order = 1
            # Shortcuts for variable names
            if variation == "l":
                variation = "lambda"
            if variation2 == "l":
                variation2 = "lambda"
            if variation == "i":
                variation = "inc"
            if variation2 == "i":
                variation2 = "inc"

            variationtypes = ["m","a","e","inc","omega","Omega","f","k","h","lambda","ix","iy"]
            if variation_order==1:
                if variation in variationtypes:
                    method = getattr(clibrebound, 'reb_derivatives_'+variation)
                    method.restype = Particle
                    p = method(c_double(simulation.G), primary, particle)
                else:
                    raise ValueError("Variational particles can only be initializes using the derivatives with respect to one of the following: %s."%", ".join(variationtypes))
            elif variation_order==2:
                if variation in variationtypes and variation2 in variationtypes:
                    # Swap variations if needed
                    vi1 = variationtypes.index(variation)
                    vi2 = variationtypes.index(variation2)
                    if vi2 < vi1:
                        variation, variation2 = variation2, variation

                    method = getattr(clibrebound, 'reb_derivatives_'+variation+'_'+variation2)
                    method.restype = Particle
                    p = method(c_double(simulation.G), primary, particle)
                else:
                    raise ValueError("Variational particles can only be initializes using the derivatives with respect to one of the following: %s."%", ".join(variationtypes))
            else:
                raise ValueError("Variational equations beyond second order are not implemented.")
            self.m = p.m
            self.x = p.x
            self.y = p.y
            self.z = p.z
            self.vx = p.vx
            self.vy = p.vy
            self.vz = p.vz
            return 

        if particle is not None:
            memmove(byref(self), byref(particle), sizeof(self))
            return
        cart = [x,y,z,vx,vy,vz]
        orbi = [primary,a,P,e,inc,Omega,omega,pomega,f,M,l,theta,T]
        pal  = [h,k,ix,iy]
       
        self.ax = 0.
        self.ay = 0.
        self.az = 0.
        if m is None:
            self.m = 0.
        else:
            self.m = m 
        if r is None:
            self.r = 0.
        else:
            self.r = r
        self.lastcollision = 0.
        self.c = None
        self.ap = None
        
        if notNone([e,inc,omega,pomega,Omega,M,f,theta,T]) and notNone(pal):
            raise ValueError("You cannot mix Pal coordinates (h,k,ix,iy) with the following orbital elements: e,inc,Omega,omega,pomega,f,M,theta,T. If a longitude/anomaly is needed in Pal coordinates, use l.")
        if notNone(cart) and notNone(orbi):
                raise ValueError("You cannot pass cartesian coordinates and orbital elements (and/or primary) at the same time.")
        if notNone(orbi):
            if simulation is None:
                raise ValueError("Need to specify simulation when initializing particle with orbital elements.")
            if primary is None:
                clibrebound.reb_get_com.restype = Particle
                primary = clibrebound.reb_get_com(byref(simulation)) # this corresponds to adding in Jacobi coordinates
            if jacobi_masses is True:
                interior_mass = 0
                for p in simulation.particles:
                    interior_mass += p.m
                # orbit conversion uses mu=G*(p.m+primary.m) so set prim.m=Mjac-m so mu=G*Mjac
                primary.m = simulation.particles[0].m*(self.m + interior_mass)/interior_mass - self.m
            if a is None and P is None:
                raise ValueError("You need to pass either a semimajor axis or orbital period to initialize the particle using orbital elements.")
            if a is not None and P is not None:
                raise ValueError("You can pass either the semimajor axis or orbital period, but not both.")
            if a is None:
                a = (P**2*simulation.G*(primary.m + self.m)/(4.*math.pi**2))**(1./3.)
            if notNone(pal):
                # Pal orbital parameters
                if h is None:
                    h = 0.
                if k is None:
                    k = 0.
                if l is None:
                    l = 0.
                if ix is None:
                    ix = 0.
                if iy is None:
                    iy = 0.
                if((ix*ix + iy*iy) > 4.0):
                    raise ValueError("Passed (ix, iy) coordinates are not valid, squared sum exceeds 4.")
                clibrebound.reb_tools_pal_to_particle.restype = Particle
                p = clibrebound.reb_tools_pal_to_particle(c_double(simulation.G), primary, c_double(self.m), c_double(a), c_double(l), c_double(k), c_double(h), c_double(ix), c_double(iy))
            else:
                # Normal orbital parameters
                if e is None:
                    e = 0.
                if inc is None:
                    inc = 0.
                if Omega is None:               # we require that Omega be passed if you want to specify longitude of node
                    Omega = 0.

                pericenters = [omega, pomega]   # Need omega for C function. Can specify it either directly or through pomega indirectly. 
                numNones = pericenters.count(None)

                if numNones == 0:
                    raise ValueError("Can't pass both omega and pomega")
                if numNones == 2:                   # Neither passed.  Default to 0.
                    omega = 0.
                if numNones == 1:
                    if pomega is not None:          # Only have to find omega is pomega was passed
                        if math.cos(inc) > 0:       # inc is in range [-pi/2, pi/2] (prograde), so pomega = Omega + omega
                            omega = pomega - Omega
                        else:
                            omega = Omega - pomega  # for retrograde orbits, pomega = Omega - omega

                longitudes = [f,M,l,theta,T]      # can specify longitude through any of these four.  Need f for C function.
                numNones = longitudes.count(None)

                if numNones < 4:
                    raise ValueError("Can only pass one longitude/anomaly in the set [f, M, l, theta, T]")
                if numNones == 5:                           # none of them passed.  Default to 0.
                    f = 0.
                if numNones == 4:                           # Only one was passed.
                    if f is None:                           # Only have to work if f wasn't passed.
                        if theta is not None:               # theta is next easiest
                            if math.cos(inc) > 0:           # for prograde orbits, theta = Omega + omega + f
                                f = theta - Omega - omega
                            else:
                                f = Omega - omega - theta   # for retrograde, theta = Omega - omega - f
                        else:                               # Either M, l, or T was passed.  Will need to find M first (if not passed) to find f
                            if l is not None:
                                if math.cos(inc) > 0:       # for prograde orbits, l = Omega + omega + M
                                    M = l - Omega - omega
                                else:
                                    M = Omega - omega - l   # for retrograde, l = Omega - omega - M
                            else:
                                if T is not None:           # works for both elliptical and hyperbolic orbits
                                                            # TODO: has accuracy problems for M=n*(t-T) << 1
                                    n = (simulation.G*(primary.m+self.m)/abs(a**3))**0.5
                                    M = n*(simulation.t - T)
                            clibrebound.reb_tools_M_to_f.restype = c_double
                            f = clibrebound.reb_tools_M_to_f(c_double(e), c_double(M))

                err = c_int()
                clibrebound.reb_tools_orbit_to_particle_err.restype = Particle
                p = clibrebound.reb_tools_orbit_to_particle_err(c_double(simulation.G), primary, c_double(self.m), c_double(a), c_double(e), c_double(inc), c_double(Omega), c_double(omega), c_double(f), byref(err))
                if err.value == 1:
                    raise ValueError("Can't set e exactly to 1.")
                if err.value == 2:
                    raise ValueError("Eccentricity must be greater than or equal to zero.")
                if err.value == 3:
                    raise ValueError("Bound orbit (a > 0) must have e < 1.")
                if err.value == 4:
                    raise ValueError("Unbound orbit (a < 0) must have e > 1.")
                if err.value == 5:
                    raise ValueError("Unbound orbit can't have f beyond the range allowed by the asymptotes set by the hyperbola.")
                if err.value == 6:
                    raise ValueError("Primary has no mass.")
            self.x = p.x
            self.y = p.y
            self.z = p.z
            self.vx = p.vx
            self.vy = p.vy
            self.vz = p.vz
        else:
            if x is None:
                x = 0.
            if y is None:
                y = 0.
            if z is None:
                z = 0.
            if vx is None:
                vx = 0.
            if vy is None:
                vy = 0.
            if vz is None:
                vz = 0.
            self.x = x
            self.y = y
            self.z = z
            self.vx = vx
            self.vy = vy
            self.vz = vz
        
    def copy(self):
        """
        Returns a deep copy of the particle. The particle is not added to any simulation by default.
        """
        np = Particle()
        memmove(byref(np), byref(self), sizeof(self))
        return np

    def calculate_orbit(self, primary=None, G=None):
        """ 
        Returns a rebound.Orbit object with the keplerian orbital elements
        corresponding to the particle around the passed primary
        (rebound.Particle) If no primary is passed, defaults to Jacobi coordinates
        (with mu = G*Minc, where Minc is the total mass from index 0 to the particle's index, inclusive). 
        
        Examples
        --------
        
        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> sim.add(x=1.,vy=1.)
        >>> orbit = sim.particles[1].calculate_orbit(sim.particles[0])
        >>> print(orbit.e) # gives the eccentricity

        Parameters
        ----------
        primary : rebound.Particle
            Central body (Optional. Default uses Jacobi coordinates)
        G : float
            Gravitational constant (Optional. Default takes G from simulation in which particle is in)
        
        Returns
        -------
        A rebound.Orbit object 
        """
        if not self._sim:
            # Particle not in a simulation
            if primary is None:
                raise ValueError("Particle does not belong to any simulation and no primary given. Cannot calculate orbit.")
            if G is None:
                raise ValueError("Particle does not belong to any simulation and G not given. Cannot calculate orbit.")
            else:
                G = c_double(G)
        else:
            # First check whether this is particles[0]
            clibrebound.reb_get_particle_index.restype = c_int
            index = clibrebound.reb_get_particle_index(byref(self)) # first check this isn't particles[0]
            if index == 0 and primary is None:
                raise ValueError("Orbital elements for particle[0] not implemented unless primary is provided")

            if primary is None:    # Use default, i.e., Jacobi coordinates
                clibrebound.reb_get_jacobi_com.restype = Particle   # now return jacobi center of mass
                primary = clibrebound.reb_get_jacobi_com(byref(self))
            G = c_double(self._sim.contents.G)
        
        err = c_int()
        clibrebound.reb_tools_particle_to_orbit_err.restype = rebound.Orbit
        o = clibrebound.reb_tools_particle_to_orbit_err(G, self, primary, byref(err))

        if err.value == 1:
            raise ValueError("Primary has no mass.")
        if err.value == 2:
            raise ValueError("Particle and primary positions are the same.")

        return o
    
    def sample_orbit(self, Npts=100, primary=None, trailing=True, timespan=None, useTrueAnomaly=None):
        """
        Returns a nested list of xyz positions along the osculating orbit of the particle. 
        If primary is not passed, returns xyz positions along the Jacobi osculating orbit
        (with mu = G*Minc, where Minc is the total mass from index 0 to the particle's index, inclusive). 

        Parameters
        ----------
        Npts    : int, optional  
            Number of points along the orbit to return  (default: 100)
        primary : rebound.Particle, optional
            Primary to use for the osculating orbit (default: Jacobi center of mass)
        trailing: bool, optional
            Whether to return points stepping backwards in time (True) or forwards (False). (default: True)
        timespan: float, optional    
            Return points (for the osculating orbit) from the current position to timespan (forwards or backwards in time depending on trailing keyword). 
            Defaults to the orbital period for bound orbits, and to the rough time it takes the orbit to move by the current distance from the primary for a hyperbolic orbit. Implementation currently only supports this option if useTrueAnomaly=False.
        useTrueAnomaly: bool, optional
            Will sample equally spaced points in true anomaly if True, otherwise in mean anomaly.
            Latter might be better for hyperbolic orbits, where true anomaly can stay near the limiting value for a long time, and then switch abruptly at pericenter. (Default: True for bound orbits, False for unbound orbits)
        """
        pts = []
        if primary is None:
            primary = self.jacobi_com
        o = self.calculate_orbit(primary=primary)

        if timespan is None:
            if o.a < 0.: # hyperbolic orbit
                timespan = 2*math.pi*o.d/o.v # rough time to cross display box
            else:
                timespan = o.P
        
        lim_phase = abs(o.n)*timespan # n is negative for hyperbolic orbits

        if trailing is True:
            lim_phase *= -1 # sample phase backwards from current value
        phase = [lim_phase*i/(Npts-1) for i in range(Npts)]

        if useTrueAnomaly is None:
            if o.a <0.:
                useTrueAnomaly = False
            else:
                useTrueAnomaly = True

        for i,ph in enumerate(phase):
            if useTrueAnomaly is True:
                newp = Particle(a=o.a, f=o.f+ph, inc=o.inc, omega=o.omega, Omega=o.Omega, e=o.e, m=self.m, primary=primary, simulation=self._sim.contents)
            else: 
                newp = Particle(a=o.a, M=o.M+ph, inc=o.inc, omega=o.omega, Omega=o.Omega, e=o.e, m=self.m, primary=primary, simulation=self._sim.contents)

            pts.append(newp.xyz)
        
        return pts

    # Simple operators for particles.
    
    def __pow__(self, other):
        if not isinstance(other, Particle):
            return NotImplemented 
        clibrebound.reb_particle_distance.restype = c_double
        return clibrebound.reb_particle_distance(byref(self), byref(other))
    def __add__(self, other):
        if not isinstance(other, Particle):
            return NotImplemented 
        c = self.copy()
        return c.__iadd__(other)
    
    def __iadd__(self, other):
        if not isinstance(other, Particle):
            return NotImplemented 
        clibrebound.reb_particle_iadd(byref(self), byref(other))
        return self
    
    def __sub__(self, other):
        if not isinstance(other, Particle):
            return NotImplemented 
        c = self.copy()
        return c.__isub__(other)
    
    def __isub__(self, other):
        if not isinstance(other, Particle):
            return NotImplemented 
        clibrebound.reb_particle_isub(byref(self), byref(other))
        return self
    
    def __mul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented 
        c = self.copy()
        return c.__imul__(other)
    
    def __imul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented 
        clibrebound.reb_particle_imul(byref(self), c_double(other))
        return self
    
    def __rmul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented 
        c = self.copy()
        return c.__imul__(other)
    
    def __div__(self, other):
        return self.__truediv__(other)
    
    def __idiv__(self, other):
        return self.__itruediv__(other)
    
    def __truediv__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented 
        c = self.copy()
        if other==0.:
            raise ZeroDivisionError
        return c.__imul__(1./other)
    
    def __itruediv__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented 
        if other==0.:
            raise ZeroDivisionError
        return self.__imul__(1./other)


    @property
    def index(self):
        clibrebound.reb_get_particle_index.restype = c_int
        return clibrebound.reb_get_particle_index(byref(self)) 
    
    @property
    def xyz(self):
        """
        Get or set the xyz position coordinates of the particle.
        """
        return [self.x, self.y, self.z]
    @xyz.setter
    def xyz(self, value):
        if len(value)!=3:
            raise AttributeError("Can only set xyz positions to array with length 3")
        self.x = float(value[0])
        self.y = float(value[1])
        self.z = float(value[2])
    
    @property
    def vxyz(self):
        """
        Get or set the xyz velocity coordinates of the particle.
        """
        return [self.vx, self.vy, self.vz]
    @vxyz.setter
    def vxyz(self, value):
        if len(value)!=3:
            raise AttributeError("Can only set xyz velocities to array with length 3")
        self.vx = float(value[0])
        self.vy = float(value[1])
        self.vz = float(value[2])
        
    @property
    def d(self):
        return self.calculate_orbit().d
    @property
    def v(self):
        return self.calculate_orbit().v 
    @property
    def h(self):
        return self.calculate_orbit().h
    @property
    def P(self):
        return self.calculate_orbit().P
    @property
    def n(self):
        return self.calculate_orbit().n 
    @property
    def a(self):
        return self.calculate_orbit().a 
    @property
    def rhill(self):
        return self.calculate_orbit().rhill
    @property
    def e(self):
        return self.calculate_orbit().e 
    @property
    def inc(self):
        return self.calculate_orbit().inc 
    @property
    def Omega(self):
        return self.calculate_orbit().Omega 
    @property
    def omega(self):
        return self.calculate_orbit().omega 
    @property
    def pomega(self):
        return self.calculate_orbit().pomega 
    @property
    def f(self):
        return self.calculate_orbit().f 
    @property
    def M(self):
        return self.calculate_orbit().M 
    @property
    def l(self):
        return self.calculate_orbit().l 
    @property
    def theta(self):
        return self.calculate_orbit().theta 
    @property
    def T(self):
        return self.calculate_orbit().T
    @property
    def orbit(self):
        return self.calculate_orbit()
    @property
    def jacobi_com(self):
        clibrebound.reb_get_jacobi_com.restype = Particle
        return clibrebound.reb_get_jacobi_com(byref(self))
    @property
    def hash(self):
        """
        Get or set the particle's hash.  If set to a string, the corresponding integer hash is calculated.
        """
        return c_uint32(self._hash)
    @hash.setter
    def hash(self, value):
        PY3 = sys.version_info[0] == 3
        hash_types = c_uint32, c_uint, c_ulong
        if PY3:
            string_types = str,
            int_types = int,
        else:
            string_types = basestring,
            int_types = int, long,
        if isinstance(value, hash_types):
            self._hash = value.value
        elif isinstance(value, string_types):
            self._hash = rebound.hash(value).value
        elif isinstance(value, int_types):
            self._hash = value
        else:
            raise AttributeError("Hash must be set to an integer, a ctypes.c_uint32 or a string. See UniquelyIdentifyingParticles.ipynb ipython_example.")



import rebound
import unittest
import os
import math
import numpy as np

class TestSerialize(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.t = 1.246
        self.sim.add(m=1.)
        self.sim.add(a=1.)
    
    def tearDown(self):
        self.sim = None

    def test_serialize(self):
        a = np.zeros((self.sim.N,3),dtype="float64")
        self.sim.serialize_particle_data(xyz=a)

        self.assertEqual(a[1][0],1)
        self.assertEqual(a[1][1],0)
        self.assertEqual(a[1][2],0)

        b = np.zeros(self.sim.N,dtype="uint32")
        c = np.zeros(self.sim.N)
        self.sim.serialize_particle_data(r=c,hash=b)

        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(r=b)

        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(hash=a)
        
        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(xyz=c)

    
if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import datetime


class TestVariational(unittest.TestCase):
    paramlist = [ 
            (1e-3,1.,0.1,0.02,0.3,0.56,0.4),
            (1e-6,2.,0.02,0.0132,0.33,1.56,0.14),
            (234.3e-6,1.7567,0.561,0.572,0.573,2.56,0.354),
            (1e-2,1.7567,0.1561,0.15472,0.24573,12.56,1.354),
            (1e-7,3.7567,0.00061,0.23572,0.523473,2.56,3.354),
            ]
    paramkeys = ["m","a","e","inc","Omega","omega","f"]
    def test_all_1st_order_full(self):
        for params in self.paramlist:
            for v in self.paramkeys:
                Delta=1e-8

                param = dict(zip(self.paramkeys, params))
                simvp = rebound.Simulation()
                simvp.add(m=1.)
                simvp.add(**param)
                simvp.add(primary=simvp.particles[0],a=1.76, m=1e-3)
                var_i = simvp.add_variation()
                var_i.vary(1,v)
                simvp.integrate(1.4)

                simsp = rebound.Simulation()
                simsp.add(m=1.)
                param[v] += Delta
                simsp.add(**param)
                simsp.add(primary=simsp.particles[0],a=1.76, m=1e-3)
                simsp.integrate(1.4)

                prec = 1e-5
                dp = (simsp.particles[1]-simvp.particles[1])/Delta - var_i.particles[1]
                self.assertLess(abs(dp.x ),prec)
                self.assertLess(abs(dp.y ),prec)
                self.assertLess(abs(dp.z ),prec)
                self.assertLess(abs(dp.vx),prec)
                self.assertLess(abs(dp.vy),prec)
                self.assertLess(abs(dp.vz),prec)
                self.assertLess(abs(dp.m ),prec)
    
    
    
    def test_all_2nd_order_full(self):
        self.run_2nd_order_full(com=False)
    def test_all_2nd_order_full_com(self):
        self.run_2nd_order_full(com=True)
    def run_2nd_order_full(self, com):
        for params in self.paramlist:
            for v1 in self.paramkeys:
                for v2 in self.paramkeys:
                    param = dict(zip(self.paramkeys, params))
                    simvp = rebound.Simulation()
                    simvp.add(m=1.)
                    simvp.add(**param)
                    simvp.add(primary=simvp.particles[0],a=1.76, m=1e-3)
                    var_ia = simvp.add_variation()
                    var_ib = simvp.add_variation()
                    var_ii = simvp.add_variation(order=2,first_order=var_ia,first_order_2=var_ib)
                    var_ia.vary(1, v1)
                    var_ib.vary(1, v2)
                    var_ii.vary(1, v1, v2)
                    if com:
                        simvp.move_to_com()
                    simvp.integrate(1.4)

                    Delta = 1e-5
                        
                    param = dict(zip(self.paramkeys, params))
                    simpp = rebound.Simulation()
                    simpp.add(m=1.)
                    param[v1] += Delta
                    param[v2] += Delta
                    simpp.add(**param)
                    simpp.add(primary=simpp.particles[0],a=1.76, m=1e-3)
                    if com:
                        simpp.move_to_com()
                    simpp.integrate(1.4)
                    
                    param = dict(zip(self.paramkeys, params))
                    simpm = rebound.Simulation()
                    simpm.add(m=1.)
                    param[v1] += Delta
                    param[v2] -= Delta
                    simpm.add(**param)
                    simpm.add(primary=simpm.particles[0],a=1.76, m=1e-3)
                    if com:
                        simpm.move_to_com()
                    simpm.integrate(1.4)

                    param = dict(zip(self.paramkeys, params))
                    simmp = rebound.Simulation()
                    simmp.add(m=1.)
                    param[v1] -= Delta
                    param[v2] += Delta
                    simmp.add(**param)
                    simmp.add(primary=simmp.particles[0],a=1.76, m=1e-3)
                    if com:
                        simmp.move_to_com()
                    simmp.integrate(1.4)

                    param = dict(zip(self.paramkeys, params))
                    simmm = rebound.Simulation()
                    simmm.add(m=1.)
                    param[v1] -= Delta
                    param[v2] -= Delta
                    simmm.add(**param)
                    simmm.add(primary=simmm.particles[0],a=1.76, m=1e-3)
                    if com:
                        simmm.move_to_com()
                    simmm.integrate(1.4)

                    prec = 1e-4
                    
                    dp = (simpp.particles[1]-simpm.particles[1]-simmp.particles[1]+simmm.particles[1])/(Delta*Delta*4.) - var_ii.particles[1]

                    self.assertLess(abs(dp.x),prec)
                    self.assertLess(abs(dp.y),prec)
                    self.assertLess(abs(dp.z),prec)
                    self.assertLess(abs(dp.vx),prec)
                    self.assertLess(abs(dp.vy),prec)
                    self.assertLess(abs(dp.vz),prec)
                    self.assertLess(abs(dp.m),prec)

class TestVariationalPal(TestVariational):
    paramkeys = ["m","a","h","k","l","ix","iy"]
    paramlist = [ 
            [1e-3,      1.,     0.,     0.,     0.,         0.0,     0.0],
            [1e-3,      1.,     0.1,    0.02,   0.3,        0.0,     0.0],
            [1e-6,      2.,     0.02,   0.0132, 0.33,       0.126,   0.14],
            [234.3e-6,  1.7567, 0.561,  0.572,  0.573,      0.056,   0.0091354],
            [1e-2,      1.7567, 0.1561, 0.15472,0.24573,    0.0056,  0.0013],
            [1e-7,      3.7567, 0.00061,0.23572,0.523473,   0.47256, 0.000024],
            [1e-7,      3.7567, 0.00061,0.23572,0.523473,   1.97,    0.0],
            [1e-7,      3.7567, 0.00061,0.23572,0.523473,   0.0,     1.97],
            ]

    
class TestVariationalTestParticle(unittest.TestCase):
    paramlist = [ 
            (1.,0.1,0.02,0.3,0.56,0.4),
            (2.,0.02,0.0132,0.33,1.56,0.14),
            (1.7567,0.561,0.572,0.573,2.56,0.354),
            (1.7567,0.1561,0.15472,0.24573,12.56,1.354),
            (3.7567,0.00061,0.23572,0.523473,2.56,3.354),
            ]
    paramkeys = ["a","e","inc","Omega","omega","f"]
    def test_all_1st_order_full(self):
        for params in self.paramlist:
            for v in self.paramkeys:
                Delta=1e-8

                param = dict(zip(self.paramkeys, params))
                simvp = rebound.Simulation()
                simvp.add(m=1.)
                simvp.add(**param)
                simvp.add(primary=simvp.particles[0],a=1.76, m=1e-3)
                var_i = simvp.add_variation(testparticle=1)
                var_i.vary(1,v)
                simvp.integrate(1.4)

                simsp = rebound.Simulation()
                simsp.add(m=1.)
                param[v] += Delta
                simsp.add(**param)
                simsp.add(primary=simsp.particles[0],a=1.76, m=1e-3)
                simsp.integrate(1.4)

                prec = 1e-5
                dp = (simsp.particles[1]-simvp.particles[1])/Delta - var_i.particles[0]
                self.assertLess(abs(dp.x ),prec)
                self.assertLess(abs(dp.y ),prec)
                self.assertLess(abs(dp.z ),prec)
                self.assertLess(abs(dp.vx),prec)
                self.assertLess(abs(dp.vy),prec)
                self.assertLess(abs(dp.vz),prec)
                self.assertLess(abs(dp.m ),prec)
    
    
    
    def test_all_2nd_order_full(self):
        self.run_2nd_order_full(com=False)
    def test_all_2nd_order_full_com(self):
        self.run_2nd_order_full(com=True)
    def run_2nd_order_full(self, com):
        for params in self.paramlist:
            for v1 in self.paramkeys:
                for v2 in self.paramkeys:
                    param = dict(zip(self.paramkeys, params))
                    simvp = rebound.Simulation()
                    simvp.add(m=1.)
                    simvp.add(**param)
                    simvp.add(primary=simvp.particles[0],a=1.76, m=1e-3)
                    var_ia = simvp.add_variation(testparticle=1)
                    var_ib = simvp.add_variation(testparticle=1)
                    var_ii = simvp.add_variation(order=2,first_order=var_ia,first_order_2=var_ib,testparticle=1)
                    var_ia.vary(1, v1)
                    var_ib.vary(1, v2)
                    var_ii.vary(1, v1, v2)
                    if com:
                        simvp.move_to_com()
                    simvp.integrate(1.4)

                    Delta = 1e-5
                        
                    param = dict(zip(self.paramkeys, params))
                    simpp = rebound.Simulation()
                    simpp.add(m=1.)
                    param[v1] += Delta
                    param[v2] += Delta
                    simpp.add(**param)
                    simpp.add(primary=simpp.particles[0],a=1.76, m=1e-3)
                    if com:
                        simpp.move_to_com()
                    simpp.integrate(1.4)
                    
                    param = dict(zip(self.paramkeys, params))
                    simpm = rebound.Simulation()
                    simpm.add(m=1.)
                    param[v1] += Delta
                    param[v2] -= Delta
                    simpm.add(**param)
                    simpm.add(primary=simpm.particles[0],a=1.76, m=1e-3)
                    if com:
                        simpm.move_to_com()
                    simpm.integrate(1.4)

                    param = dict(zip(self.paramkeys, params))
                    simmp = rebound.Simulation()
                    simmp.add(m=1.)
                    param[v1] -= Delta
                    param[v2] += Delta
                    simmp.add(**param)
                    simmp.add(primary=simmp.particles[0],a=1.76, m=1e-3)
                    if com:
                        simmp.move_to_com()
                    simmp.integrate(1.4)

                    param = dict(zip(self.paramkeys, params))
                    simmm = rebound.Simulation()
                    simmm.add(m=1.)
                    param[v1] -= Delta
                    param[v2] -= Delta
                    simmm.add(**param)
                    simmm.add(primary=simmm.particles[0],a=1.76, m=1e-3)
                    if com:
                        simmm.move_to_com()
                    simmm.integrate(1.4)

                    prec = 1e-4
                    
                    dp = (simpp.particles[1]-simpm.particles[1]-simmp.particles[1]+simmm.particles[1])/(Delta*Delta*4.) - var_ii.particles[0]

                    self.assertLess(abs(dp.x),prec)
                    self.assertLess(abs(dp.y),prec)
                    self.assertLess(abs(dp.z),prec)
                    self.assertLess(abs(dp.vx),prec)
                    self.assertLess(abs(dp.vy),prec)
                    self.assertLess(abs(dp.vz),prec)
                    self.assertLess(abs(dp.m),prec)

class TestVariationalPalTestParticle(TestVariationalTestParticle):
    paramkeys = ["a","h","k","l","ix","iy"]
    paramlist = [ 
            [1.,     0.,     0.,     0.,         0.0,     0.0],
            [1.,     0.1,    0.02,   0.3,        0.0,     0.0],
            [2.,     0.02,   0.0132, 0.33,       0.126,   0.14],
            [1.7567, 0.561,  0.572,  0.573,      0.056,   0.0091354],
            [1.7567, 0.1561, 0.15472,0.24573,    0.0056,  0.0013],
            [3.7567, 0.00061,0.23572,0.523473,   0.47256, 0.000024],
            [3.7567, 0.00061,0.23572,0.523473,   1.97,    0.0],
            [3.7567, 0.00061,0.23572,0.523473,   0.0,     1.97],
            ]

if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import os
import rebound.data as data

class TestData(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
    
    def tearDown(self):
        self.sim = None
    
    def test_add_outer_solar_system(self):
        data.add_outer_solar_system(self.sim)
        self.assertEqual(self.sim.N,6)
        self.assertAlmostEqual(self.sim.particles[5].x,-21.3858977531573,delta=1e-15)
        self.assertAlmostEqual(self.sim.particles[5].a,39.485206935092286,delta=1e-15)




import rebound
import unittest
import rebound.data as data

class TestAdditionalForces(unittest.TestCase):
    def test_af_ias15(self):
        sim = rebound.Simulation()
        sim.integrator = "ias15"
        sim.force_is_velocity_dependent = 1
        sim.add(m=1)
        sim.add(m=1e-6,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        def af(sim):
            fac = 0.01
            sim.contents.particles[2].ax -= fac*sim.contents.particles[2].vx 
            sim.contents.particles[2].ay -= fac*sim.contents.particles[2].vz 
            sim.contents.particles[2].az -= fac*sim.contents.particles[2].vy 
        sim.additional_forces = af
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].a,4.86583,delta=1e-5)

    def test_af_mercurius(self):
        sim = rebound.Simulation()
        sim.integrator = "mercurius"
        sim.dt = 0.005
        sim.force_is_velocity_dependent = 1
        sim.add(m=1)
        sim.add(m=1e-6,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        def af(sim):
            fac = 0.01
            sim.contents.particles[2].ax -= fac*sim.contents.particles[2].vx 
            sim.contents.particles[2].ay -= fac*sim.contents.particles[2].vz 
            sim.contents.particles[2].az -= fac*sim.contents.particles[2].vy 
        sim.additional_forces = af
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].a,4.86583,delta=1e-4)
    
    def test_af_mercurius_closeencounter(self):
        sim = rebound.Simulation()
        sim.integrator = "mercurius"
        sim.ri_mercurius.hillfac = 100000 # make sure encounter happens
        sim.dt = 0.005
        sim.force_is_velocity_dependent = 1
        sim.add(m=1)
        sim.add(m=1e-6,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        def af(sim):
            fac = 0.01
            sim.contents.particles[2].ax -= fac*sim.contents.particles[2].vx 
            sim.contents.particles[2].ay -= fac*sim.contents.particles[2].vz 
            sim.contents.particles[2].az -= fac*sim.contents.particles[2].vy 
        sim.additional_forces = af
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].a,4.86583,delta=1e-4)


    def test_af_whfast(self):
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.dt = 0.01
        sim.force_is_velocity_dependent = 1
        sim.add(m=1)
        sim.add(m=1e-6,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        def af(sim):
            fac = 0.01
            sim.contents.particles[2].ax -= fac*sim.contents.particles[2].vx 
            sim.contents.particles[2].ay -= fac*sim.contents.particles[2].vz 
            sim.contents.particles[2].az -= fac*sim.contents.particles[2].vy 
        sim.additional_forces = af
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].a,4.86583,delta=1e-4)

    def test_af_whfastjac(self):
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "jacobi"
        sim.dt = 0.01
        sim.force_is_velocity_dependent = 1
        sim.add(m=1)
        sim.add(m=1e-6,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        def af(sim):
            fac = 0.01
            sim.contents.particles[2].ax -= fac*sim.contents.particles[2].vx 
            sim.contents.particles[2].ay -= fac*sim.contents.particles[2].vz 
            sim.contents.particles[2].az -= fac*sim.contents.particles[2].vy 
        sim.additional_forces = af
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].a,4.86583,delta=1e-4)


if __name__ == "__main__":
    unittest.main()




import rebound
import unittest
import rebound.data as data

mdot = 1e-6
def ptm(sim):
    if sim.contents.ri_mercurius.mode == 0:
        sim.contents.particles[2].m -= sim.contents.dt_last_done*mdot

class TestPostTimestepModifications(unittest.TestCase):

    def test_ptm_ias15(self):
        sim = rebound.Simulation()
        sim.integrator = "ias15"
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        sim.post_timestep_modifications = ptm
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].m,1e-3-mdot*sim.t,delta=1e-13)

    def test_ptm_mercurius(self):
        sim = rebound.Simulation()
        sim.integrator = "mercurius"
        sim.dt = 0.01
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        sim.post_timestep_modifications = ptm
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].m,1e-3-mdot*sim.t,delta=1e-13)
    
    def test_ptm_mercurius_closeencounter(self):
        sim = rebound.Simulation()
        sim.integrator = "mercurius"
        sim.dt = 0.01
        sim.ri_mercurius.hillfac = 100000 # make sure encounter happens
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        sim.post_timestep_modifications = ptm
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].m,1e-3-mdot*sim.t,delta=1e-13)

    def test_ptm_whfast(self):
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.dt = 0.01
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        sim.post_timestep_modifications = ptm
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].m,1e-3-mdot*sim.t,delta=1e-13)

    def test_ptm_whfastjac(self):
        sim = rebound.Simulation()
        sim.integrator = "whfast"
        sim.ri_whfast.coordinates = "jacobi"
        sim.dt = 0.01
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=5)
        sim.move_to_com()
        sim.post_timestep_modifications = ptm
        sim.integrate(10.)
        self.assertAlmostEqual(sim.particles[2].m,1e-3-mdot*sim.t,delta=1e-13)


if __name__ == "__main__":
    unittest.main()




import rebound
import unittest
import warnings

class TestSimulationArchiveMatrix(unittest.TestCase):
    pass

def runSimulation(tmax=40., restart=False, keep_unsynchronized=1, interval=None, safe_mode=True, integrator="ias15",G=1., testparticle=0,simulationarchive_version=2):
    if restart:
        if keep_unsynchronized==1:
            sim = rebound.Simulation("test.bin")
        else:
            sa = rebound.SimulationArchive("test.bin")
            sim = sa.getSimulation(sa.tmax,keep_unsynchronized=0)
    else:
        sim = rebound.Simulation()
        sim.G = G
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.add(m=1e-3,a=-2,e=1.1,omega=0.1,M=0.1,inc=0.1,Omega=0.1)
        sim.integrator = integrator
        sim.dt = 0.1313
        if simulationarchive_version==1:
            sim.simulationarchive_version = 1
        if safe_mode==False:
            sim.ri_whfast.safe_mode = 1
            sim.ri_mercurius.safe_mode = 1
        if testparticle>0:
            if testparticle==1:
                sim.testparticle_type=0
            if testparticle==2:
                sim.testparticle_type=1
            sim.add(m=1e-4,a=1.2,e=0.04,omega=0.21,M=1.41,inc=0.21,Omega=1.1)
            sim.N_active = sim.N-1 # one test particle
        if interval:
            sim.automateSimulationArchive("test.bin", interval, deletefile=True)
    sim.integrate(tmax,exact_finish_time=0)
    return sim

def compareSim(test,sim1,sim2):
    test.assertEqual(sim1.N,sim2.N)
    test.assertEqual(sim1.N_active,sim2.N_active)
    test.assertEqual(sim1.N_var,sim2.N_var)
    test.assertEqual(sim1.t,sim2.t)
    test.assertEqual(sim1.G,sim2.G)
    for i in range(sim1.N):
        test.assertEqual(sim1.particles[i].r,sim2.particles[i].r)
        test.assertEqual(sim1.particles[i].m,sim2.particles[i].m)
        test.assertEqual(sim1.particles[i].x,sim2.particles[i].x)
        test.assertEqual(sim1.particles[i].y,sim2.particles[i].y)
        test.assertEqual(sim1.particles[i].z,sim2.particles[i].z)
        test.assertEqual(sim1.particles[i].vx,sim2.particles[i].vx)
        test.assertEqual(sim1.particles[i].vy,sim2.particles[i].vy)
        test.assertEqual(sim1.particles[i].vz,sim2.particles[i].vz)

def create_test_sa_restart(params):
    def doTest(self): 
        runSimulation(40., restart=False, interval=10., **params)
        sim1 = runSimulation(80., restart=True, **params)
        sim2 = runSimulation(80., restart=False, **params)
        compareSim(self,sim1,sim2)
    return doTest

def create_test_sa_synchronize(params):
    def doTest2(self): 
        sim1 = runSimulation(40., restart=False, interval=10., **params)
        if params['keep_unsynchronized']==1:
            sim2 = rebound.Simulation("test.bin")
        else:
            sa = rebound.SimulationArchive("test.bin")
            sim2 = sa.getSimulation(sa.tmax,keep_unsynchronized=0)
        compareSim(self,sim1,sim2)
        sim1.integrate(sim1.t+12.)
        sim2.integrate(sim2.t+12.)
        compareSim(self,sim1,sim2)
    return doTest2



for integrator in ["ias15","whfast","leapfrog","janus","mercurius","saba","sabacl4", "saba(10,6,4)"]:
    for safe_mode in [True,False]:
        for G in [1.,0.9]:
            for testparticle in [0,1,2]: # no test particle, passive, semi-active
                for keep_unsynchronized in [1,0]:
                    for simulationarchive_version in [1,2]:
                        if simulationarchive_version==1:
                            if integrator=="leapfrog" or integrator=="saba" or integrator=="mercurius" or integrator=="sabacl4" or integrator=="saba(10,6,4)":
                                continue
                        params = {'safe_mode':safe_mode,
                                'integrator':integrator,
                                'G':G, 
                                'testparticle':testparticle,
                                'keep_unsynchronized':keep_unsynchronized,
                                'simulationarchive_version':simulationarchive_version}
                        test_method = create_test_sa_restart(params)
                        name = "test_sa_restart"
                        for key in params:
                            name += "_"+key+":"+str(params[key])
                        test_method.__name__ = name
                        setattr(TestSimulationArchiveMatrix, name,test_method)
                        
                        test_method = create_test_sa_synchronize(params)
                        name = "test_sa_synchronize"
                        for key in params:
                            name += "_"+key+":"+str(params[key])
                        test_method.__name__ = name
                        setattr(TestSimulationArchiveMatrix, name,test_method)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""An N-body integrator package for python."""
# Make changes for python 2 and 3 compatibility
try:
    import builtins      # if this succeeds it's python 3.x
    builtins.xrange = range
    builtins.basestring = (str,bytes)
except ImportError:
    pass                 # python 2.x

# Find suffix
import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

# Import shared library
import os
import warnings
pymodulepath = os.path.dirname(__file__)
from ctypes import cdll, c_char_p
clibrebound = cdll.LoadLibrary(pymodulepath+"/../librebound"+suffix)

# Version
__version__ = c_char_p.in_dll(clibrebound, "reb_version_str").value.decode('ascii')

# Build
__build__ = c_char_p.in_dll(clibrebound, "reb_build_str").value.decode('ascii')

# Githash
__githash__ = c_char_p.in_dll(clibrebound, "reb_githash_str").value.decode('ascii')

# Check for version
try:
    moduleversion = pkg_resources.require("rebound")[0].version
    libreboundversion = __version__
    if moduleversion != libreboundversion:
        warnings.warn("WARNING: python module and librebound have different version numbers: '%s' vs '%s'.\n" %(moduleversion, libreboundversion), ImportWarning)
except:
    # Might fails on python3 versions, but not important
    pass

# Exceptions    
class SimulationError(Exception):  
    """The simulation exited with a generic error."""
    pass

class Encounter(Exception):
    """The simulation exited because a close encounter has been detected.
    You may want to search for the pair of bodies which have the smallest distance."""
    pass

class Collision(Exception):
    """The simulation exited because a collision has been detected.
    You may want to search for which particles have a lastcollision time equal to the simulation time."""
    pass

class Escape(Exception):
    """The simulation exited because a particle has been se encounter has been detected.
    You may want to search for the particle with the largest distance from the 
    origin and remove it from the simulation."""
    pass

class NoParticles(Exception):
    """The simulation exited because no particles are left in the simulation."""
    pass
    
class ParticleNotFound(Exception):
    """Particle was not found in the simulation."""
    pass

from .simulation import Simulation, Orbit, Variation, reb_simulation_integrator_saba, reb_simulation_integrator_whfast, reb_simulation_integrator_sei, reb_simulation_integrator_mercurius
from .particle import Particle
from .plotting import OrbitPlot
from .tools import hash
from .simulationarchive import SimulationArchive
from .interruptible_pool import InterruptiblePool

__all__ = ["__version__", "__build__", "__githash__", "SimulationArchive", "Simulation", "Orbit", "OrbitPlot", "Particle", "SimulationError", "Encounter", "Collision", "Escape", "NoParticles", "ParticleNotFound", "InterruptiblePool","Variation", "reb_simulation_integrator_whfast", "reb_simulation_integrator_ias15", "reb_simulation_integrator_saba", "reb_simulation_integrator_sei","reb_simulation_integrator_mercurius", "clibrebound"]



shader_code = """
<script id="orbit_shader-vs" type="x-shader/x-vertex">
    uniform vec3 focus;
    uniform vec3 aef;
    uniform vec3 omegaOmegainc;
    attribute float lintwopi;
    varying float lin;
    uniform mat4 mvp;
    const float M_PI = 3.14159265359;
    void main() {
       float a = aef.x;
       float e = aef.y;
       float f = aef.z+lintwopi;
       lin = lintwopi/(M_PI*2.);
       if (e>1.){
           float theta_max = acos(-1./e);
           f = 0.0001-theta_max+1.9998*lin*theta_max;
           lin = sqrt(min(0.5,lin));
       }
       float omega = omegaOmegainc.x;
       float Omega = omegaOmegainc.y;
       float inc = omegaOmegainc.z;
       float r = a*(1.-e*e)/(1. + e*cos(f));
       float cO = cos(Omega);
       float sO = sin(Omega);
       float co = cos(omega);
       float so = sin(omega);
       float cf = cos(f);
       float sf = sin(f);
       float ci = cos(inc);
       float si = sin(inc);
       vec3 pos = vec3(r*(cO*(co*cf-so*sf) - sO*(so*cf+co*sf)*ci),r*(sO*(co*cf-so*sf) + cO*(so*cf+co*sf)*ci),+ r*(so*cf+co*sf)*si);
       gl_Position = mvp*(vec4(focus+pos, 1.0));
    }
</script>
<script id="orbit_shader-fs" type="x-shader/x-fragment">
    precision mediump float;
    varying float lin;
    void main() {
      float fog = max(max(0.,-1.+2.*gl_FragCoord.z),max(0.,1.-2.*gl_FragCoord.z));
      gl_FragColor = vec4(1.,1.,1.,sqrt(lin)*(1.-fog));
    }
</script>
<script id="point_shader-vs" type="x-shader/x-vertex">
    attribute vec3 vp;
    uniform mat4 mvp;
    //uniform vec4 vc;
    //varying vec4 color;
    void main() {
      gl_PointSize = 15.0;
      gl_Position = mvp*vec4(vp, 1.0);
      //color = vc;
    }
</script>
<script id="point_shader-fs" type="x-shader/x-fragment">
    precision mediump float;
    //varying vec4 color;
    void main() {
      vec2 rel = gl_PointCoord.st;
      rel.s -=0.5;
      rel.t -=0.5;
      if (length(rel)>0.25){
         gl_FragColor = vec4(0.,0.,0.,0.); 
      }else{
         vec4 cmod = vec4(1.,1.,1.,1.);
         float fog = max(max(0.,-1.+2.*gl_FragCoord.z),max(0.,1.-2.*gl_FragCoord.z));
         cmod.a*= (1.-fog)*min(1.,1.-4.*(length(rel)/0.25-0.75));
         gl_FragColor = cmod;
      }
    }
</script>
"""
js_code = """
<script>
function compileShader(glr, shaderSource, shaderType) {
  // Create the shader object
  var shader = glr.createShader(shaderType);
 
  // Set the shader source code.
  glr.shaderSource(shader, shaderSource);
 
  // Compile the shader
  glr.compileShader(shader);
 
  // Check if it compiled
  var success = glr.getShaderParameter(shader, glr.COMPILE_STATUS);
  if (!success) {
    // Something went wrong during compilation; get the error
    throw "could not compile shader:" + glr.getShaderInfoLog(shader);
  }
 
  return shader;
}
function createShaderFromScript(glr, scriptId, opt_shaderType) {
  // look up the script tag by id.
  var shaderScript = document.getElementById(scriptId);
  if (!shaderScript) {
    throw("*** Error: unknown script element" + scriptId);
  }
 
  // extract the contents of the script tag.
  var shaderSource = shaderScript.text;
 
  // If we didn't pass in a type, use the 'type' from
  // the script tag.
  if (!opt_shaderType) {
    if (shaderScript.type == "x-shader/x-vertex") {
      opt_shaderType = glr.VERTEX_SHADER;
    } else if (shaderScript.type == "x-shader/x-fragment") {
      opt_shaderType = glr.FRAGMENT_SHADER;
    } else if (!opt_shaderType) {
      throw("*** Error: shader type not set");
    }
  }
 
  return compileShader(glr, shaderSource, opt_shaderType);
};
function createProgramFromScripts( glr, vertexShaderId, fragmentShaderId) {
  var vertexShader = createShaderFromScript(glr, vertexShaderId, glr.VERTEX_SHADER);
  var fragmentShader = createShaderFromScript(glr, fragmentShaderId, glr.FRAGMENT_SHADER);
  var program = glr.createProgram();
 
  // attach the shaders.
  glr.attachShader(program, vertexShader);
  glr.attachShader(program, fragmentShader);
 
  // link the program.
  glr.linkProgram(program);
 
  // Check if it linked.
  var success = glr.getProgramParameter(program, glr.LINK_STATUS);
  if (!success) {
      // something went wrong with the link
      throw ("program filed to link:" + glr.getProgramInfoLog (program));
  }
 
  return program;
}
function quat2mat(A,mat){
    var xx = A.x*A.x; var xy = A.x*A.y; var xz = A.x*A.z;
    var xw = A.x*A.w; var yy = A.y*A.y; var yz = A.y*A.z;
    var yw = A.y*A.w; var zz = A.z*A.z; var zw = A.z*A.w;
    mat[0] = 1.-2.*(yy+zz);
    mat[1] =    2.*(xy-zw);
    mat[2] =    2.*(xz+yw);
    mat[4] =    2.*(xy+zw);
    mat[5] = 1.-2.*(xx+zz);
    mat[6] =    2.*(yz-xw);
    mat[8] =    2.*(xz-yw);
    mat[9] =    2.*(yz+xw);
    mat[10]= 1.-2.*(xx+yy);
    mat[3] = mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.; mat[15]= 1.;
}
function multvec(A, B, vecr){
    var mat = [0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.];
    quat2mat(A,mat);
    vecr[0] = mat[0]*B[0] + mat[1]*B[1] + mat[2]*B[2];
    vecr[1] = mat[4]*B[0] + mat[5]*B[1] + mat[6]*B[2];
    vecr[2] = mat[8]*B[0] + mat[9]*B[1] + mat[10]*B[2];
}
function mattransp(mat){
    var matt = [
        mat[0], mat[4], mat[8], mat[12],
        mat[1], mat[5], mat[9], mat[13],
        mat[2], mat[6], mat[10], mat[14],
        mat[3], mat[7], mat[11], mat[15]];
    return matt;
}
function conjugate(quat){
    var cquat = {x:-quat.x, y:-quat.y, z:-quat.z, w:quat.w};
    return cquat;
}
function mult(A, B){
    var mquat = {   x: A.w*B.x + A.x*B.w + A.y*B.z - A.z*B.y,
                    y: A.w*B.y - A.x*B.z + A.y*B.w + A.z*B.x,
                    z: A.w*B.z + A.x*B.y - A.y*B.x + A.z*B.w,
                    w: A.w*B.w - A.x*B.x - A.y*B.y - A.z*B.z};
    return mquat;
}

function normalize(quat){
    var L = Math.sqrt(quat.x*quat.x + quat.y*quat.y + quat.z*quat.z + quat.w*quat.w);
    var nquat = {x:quat.x/L, y:quat.y/L, z:quat.z/L, w:quat.w/L};
    return nquat;
}
function matortho(mat, l, r, b, t, n, f){
    mat[0] = 2./(r-l); mat[1] = 0.; mat[2] = 0.; mat[3] = -(r+l)/(r-l);
    mat[4] = 0.; mat[5] = 2./(t-b); mat[6] = 0.; mat[7] = -(t+b)/(t-b);
    mat[8] = 0.; mat[9] = 0.; mat[10] = -2./(f-n); mat[11] = -(f+n)/(f-n);
    mat[12] = 0.; mat[13] = 0.; mat[14] = 0.; mat[15] = 1.;
}
function matmult(A,B,C){
    for(i=0;i<4;i++){
    for(j=0;j<4;j++){
        C[i+4*j] = 0.;
    for(k=0;k<4;k++){
        C[i+4*j] += A[k+4*j]*B[i+4*k];
    }}}
}
function startGL(reboundView) {
    var canvas = document.getElementById("reboundcanvas-"+reboundView.cid);
    if (!canvas){
        reboundView.startCount = reboundView.startCount+1;
        if (reboundView.startCount>1000){
            console.log("Cannot find element.");
        }else{
            setTimeout(function(){ startGL(reboundView); }, 10);
        }
        return;
    }
    var rect = canvas.getBoundingClientRect()
    reboundView.ratio = rect.width/rect.height;
    reboundView.view = normalize({x:reboundView.orientation[0], y:reboundView.orientation[1], z:reboundView.orientation[2], w:reboundView.orientation[3]});

    canvas.addEventListener('mousedown', function() {
        reboundView.mouseDown=1;
        }, false);
    canvas.addEventListener('mouseup', function() {
        reboundView.mouseDown=0;
        }, false);
    canvas.addEventListener('mouseleave', function() {
        reboundView.mouseDown=0;
        }, false);

    canvas.addEventListener('mousemove', function(evt) {
        var rect = canvas.getBoundingClientRect()
        if (reboundView.mouseDown==1){
            reboundView.mouseDown = 2;
            reboundView.mouse_x = evt.clientX-rect.left;
            reboundView.mouse_y = evt.clientY-rect.top;
            return;
        }else if (reboundView.mouseDown==2){
            var width = rect.width;
            var height = rect.height;
            var dx = 3.*(evt.clientX-rect.left-reboundView.mouse_x)/width;
            var dy = 3.*(evt.clientY-rect.top-reboundView.mouse_y)/height;
            reboundView.mouse_x = evt.clientX-rect.left;
            reboundView.mouse_y = evt.clientY-rect.top;
            if (evt.shiftKey){
                reboundView.scale *= (1.+dx+dy);
            }else{
                var inv = conjugate(reboundView.view);
                var up = [0.,1.,0.];
                var right = [1.,0.,0.];
                var inv_up = [0.,0.,0.];
                var inv_right = [0.,0.,0.];
                multvec(inv, right, inv_right);
                multvec(inv, up, inv_up);
                
                var sin_dy = Math.sin(dy);
                var rot_dy = {x:inv_right[0]*sin_dy, y:inv_right[1]*sin_dy, z:inv_right[2]*sin_dy, w:Math.cos(dy)};
                reboundView.view = mult(reboundView.view, normalize(rot_dy));
                
                var sin_dx = Math.sin(dx);
                var rot_dx = {x:inv_up[0]*sin_dx, y:inv_up[1]*sin_dx, z:inv_up[2]*sin_dx, w:Math.cos(dx)};
                reboundView.view = normalize(mult(reboundView.view, normalize(rot_dx)));
            }

            drawGL(reboundView);
        }


        }, false);

    reboundView.gl = canvas.getContext("webgl")||canvas.getContext("experimental-webgl");
    if (!reboundView.gl) {
      alert("Unable to initialize WebGL. Your browser may not support it.");
      return;
    }
    var gl = reboundView.gl
    gl.enable(gl.BLEND);
    gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
    
    reboundView.orbit_shader_program = createProgramFromScripts(gl,"orbit_shader-vs","orbit_shader-fs");
    reboundView.point_shader_program = createProgramFromScripts(gl,"point_shader-vs","point_shader-fs");
   
    var lintwopi = new Float32Array(500);
    for(i=0;i<500;i++){
        lintwopi[i] = 2.*Math.PI/500.*i;
    }
    reboundView.orbit_lintwopi_buffer = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, reboundView.orbit_lintwopi_buffer);
    gl.bufferData(gl.ARRAY_BUFFER, 4*500, gl.STATIC_DRAW);
    gl.bufferSubData(gl.ARRAY_BUFFER, 0, lintwopi)
    reboundView.orbit_shader_mvp_location = gl.getUniformLocation(reboundView.orbit_shader_program,"mvp");
    reboundView.orbit_shader_focus_location = gl.getUniformLocation(reboundView.orbit_shader_program,"focus");
    reboundView.orbit_shader_aef_location = gl.getUniformLocation(reboundView.orbit_shader_program,"aef");
    reboundView.orbit_shader_omegaOmegainc_location = gl.getUniformLocation(reboundView.orbit_shader_program,"omegaOmegainc");
    
    reboundView.particle_data_buffer = gl.createBuffer();
    gl.useProgram(reboundView.point_shader_program);
    reboundView.point_shader_mvp_location = gl.getUniformLocation(reboundView.point_shader_program,"mvp");
    
    updateRenderData(reboundView);
    gl.clearColor(0.0, 0.0, 0.0, 1.0);
    gl.clear(gl.COLOR_BUFFER_BIT);
    drawGL(reboundView);
}
function updateRenderData(reboundView){
    var overlay = document.getElementById("reboundoverlay-"+reboundView.cid);
    overlay.innerHTML = reboundView.model.get("overlay");
    var previousN = reboundView.N;
    reboundView.N = reboundView.model.get("N");
    reboundView.t = reboundView.model.get("t");
    reboundView.particle_data = reboundView.model.get('particle_data');
    if (reboundView.orbits){
        reboundView.orbit_data = reboundView.model.get('orbit_data');
    }
    var gl = reboundView.gl
    if (reboundView.N>0){
        gl.bindBuffer(gl.ARRAY_BUFFER, reboundView.particle_data_buffer);
        gl.bufferData(gl.ARRAY_BUFFER, reboundView.N*7*4, gl.DYNAMIC_DRAW);
        gl.bufferSubData(gl.ARRAY_BUFFER, 0, reboundView.particle_data)
    }
}
function drawGL(reboundView) {
    if (!reboundView.gl){
        return;
    }
    // Cleanup
    var gl = reboundView.gl
    gl.clearColor(0.0, 0.0, 0.0, 1.0);
    gl.clear(gl.COLOR_BUFFER_BIT);
    
    // Draw
    gl.useProgram(reboundView.point_shader_program);
    gl.bindBuffer(gl.ARRAY_BUFFER, reboundView.particle_data_buffer);
    var pvp = gl.getAttribLocation(reboundView.point_shader_program,"vp");
    gl.enableVertexAttribArray(pvp);
    gl.vertexAttribPointer(pvp, 3, gl.FLOAT, 0, 4*7,0); // 4 = size of float
    var projection = [0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.];
    if (reboundView.ratio>=1.){
        matortho(projection, 
                -1.6*reboundView.scale, 1.6*reboundView.scale,
                -1.6/reboundView.ratio*reboundView.scale, 1.6/reboundView.ratio*reboundView.scale,
                -2.5*reboundView.scale, 2.5*reboundView.scale);
    }else{
        matortho(projection, 
                -1.6*reboundView.ratio*reboundView.scale, 1.6*reboundView.ratio*reboundView.scale,
                -1.6*reboundView.scale, 1.6*reboundView.scale,
                -2.5*reboundView.scale, 2.5*reboundView.scale);
    }
    var view = [0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.];
    quat2mat(reboundView.view,view);
    var mvp = [0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.];
    matmult(projection,view,mvp);
    gl.uniformMatrix4fv(reboundView.point_shader_mvp_location,false,mattransp(mvp));
    gl.drawArrays(gl.POINTS,0,reboundView.N);
   
    if (reboundView.orbits){
        gl.useProgram(reboundView.orbit_shader_program);
        gl.bindBuffer(gl.ARRAY_BUFFER, reboundView.orbit_lintwopi_buffer);
        var ltp = gl.getAttribLocation(reboundView.orbit_shader_program,"lintwopi");
        gl.enableVertexAttribArray(ltp);
        gl.vertexAttribPointer(ltp, 1, gl.FLOAT, 0, 0,0); // 4 = size of float
        gl.uniformMatrix4fv(reboundView.orbit_shader_mvp_location,false,mattransp(mvp));

        // Need to do this one by one
        // because WebGL is not supporting
        // instancing:
        for(i=0;i<reboundView.N-1;i++){
            var focus = new Float32Array(reboundView.orbit_data.buffer,4*9*i,3);
            gl.uniform3fv(reboundView.orbit_shader_focus_location,focus);
            var aef = new Float32Array(reboundView.orbit_data.buffer,4*(9*i+3),3);
            gl.uniform3fv(reboundView.orbit_shader_aef_location,aef);
            var omegaOmegainc = new Float32Array(reboundView.orbit_data.buffer,4*(9*i+6),3);
            gl.uniform3fv(reboundView.orbit_shader_omegaOmegainc_location,omegaOmegainc);

            gl.drawArrays(gl.LINE_STRIP,0,500);
        }
    }
}
require.undef('rebound');
    define('rebound', ["@jupyter-widgets/base"], function(widgets) {
    var ReboundView = widgets.DOMWidgetView.extend({
        render: function() {
            this.el.innerHTML = '<span style="display: inline-block; position: relative;" width="'+this.model.get("width")+'" height="'+this.model.get("height")+'"><canvas style="border: none;" id="reboundcanvas-'+this.cid+'" width="'+this.model.get("width")+'" height="'+this.model.get("height")+'"></canvas><span style="position: absolute; color: #FFF; pointer-events:none;  bottom:5px; right:0px; padding-right:5px; font-family: monospace;" id="reboundoverlay-'+this.cid+'">REBOUND</span></span>';
            this.model.on('change:t', this.trigger_refresh, this);
            this.model.on('change:count', this.trigger_refresh, this);
            this.model.on('change:screenshotcount', this.take_screenshot, this);
            this.startCount = 0;
            this.gl = null;
            // Only copy those once
            this.scale = this.model.get("scale");
            this.width = this.model.get("width");
            this.height = this.model.get("height");
            this.orbits = this.model.get("orbits");
            this.orientation = this.model.get("orientation");
            startGL(this);
        },
        take_screenshot: function() {
            drawGL(this);
            var canvas = document.getElementById("reboundcanvas-"+this.cid);
            var img = canvas.toDataURL("image/png");
            this.model.set("screenshot",img, {updated_view: this});
            this.touch();
        },
        trigger_refresh: function() {
            updateRenderData(this);
            drawGL(this);
        },
    });
    return {
        ReboundView: ReboundView
    };
});
      
</script>
"""

import ipywidgets
ipywidgets_major_version = int((ipywidgets.__version__).split(".")[0])


if ipywidgets_major_version<7:
    js_code = js_code.replace("@jupyter-widgets/base", "jupyter-js-widgets")
    js_code = js_code.replace(".cid", ".id")

from ipywidgets import DOMWidget
import traitlets
import math
import base64
import sys
from ctypes import c_float, byref, create_string_buffer, c_int, c_char, pointer
from . import clibrebound
def savescreenshot(change):
    if len(change["new"]) and change["type"] =="change":
        w = change["owner"]
        bd = base64.b64decode(change["new"].split(",")[-1])
        if sys.version_info[0] < 3:
            with open(w.screenshotprefix+"%05d.png"%w.screenshotcountall, 'w') as f:
                f.write(bd)
        else:
            with open(w.screenshotprefix+"%05d.png"%w.screenshotcountall, 'bw') as f:
                f.write(bd)
        w.screenshotcountall += 1
        if len(w.times)>w.screenshotcount:
            nexttime = w.times[w.screenshotcount]
            if w.archive:
                sim = w.archive.getSimulation(w.times[w.screenshotcount],mode=w.mode)
                w.refresh(pointer(sim))
            else:
                w.simp.contents.integrate(w.times[w.screenshotcount])
            w.screenshotcount += 1
        else:
            w.unobserve(savescreenshot)
            w.times = None
            w.screenshotprefix = None

class Widget(DOMWidget):
    _view_name = traitlets.Unicode('ReboundView').tag(sync=True)
    _view_module = traitlets.Unicode('rebound').tag(sync=True)
    count = traitlets.Int(0).tag(sync=True)
    screenshotcount = traitlets.Int(0).tag(sync=True)
    t = traitlets.Float().tag(sync=True)
    N = traitlets.Int().tag(sync=True)
    overlay = traitlets.Unicode('REB WIdund').tag(sync=True)
    width = traitlets.Float().tag(sync=True)
    height = traitlets.Float().tag(sync=True)
    scale = traitlets.Float().tag(sync=True)
    particle_data = traitlets.CBytes(allow_none=True).tag(sync=True)
    orbit_data = traitlets.CBytes(allow_none=True).tag(sync=True)
    orientation = traitlets.Tuple().tag(sync=True)
    orbits = traitlets.Int().tag(sync=True)
    screenshot = traitlets.Unicode().tag(sync=True)
    def __init__(self,simulation,size=(200,200),orientation=(0.,0.,0.,1.),scale=None,autorefresh=True,orbits=True, overlay=True):
        """ 
        Initializes a Widget.

        Widgets provide real-time 3D interactive visualizations for REBOUND simulations 
        within Jupyter Notebooks. To use widgets, the ipywidgets package needs to be installed
        and enabled in your Jupyter notebook server. 

        Parameters
        ----------
        size : (int, int), optional
            Specify the size of the widget in pixels. The default is 200 times 200 pixels.
        orientation : (float, float, float, float), optional
            Specify the initial orientation of the view. The four floats correspond to the
            x, y, z, and w components of a quaternion. The quaternion will be normalized.
        scale : float, optional
            Set the initial scale of the view. If not set, the widget will determine the 
            scale automatically based on current particle positions.
        autorefresh : bool, optional
            The default value if True. The view is updated whenever a particle is added,
            removed and every 100th of a second while a simulation is running. If set 
            to False, then the user needs to manually call the refresh() function on the 
            widget. This might be useful if performance is an issue.
        orbits : bool, optional
            The default value for this is True and the widget will draw the instantaneous 
            orbits of the particles. For simulations in which particles are not on
            Keplerian orbits, the orbits shown will not be accurate. 
        overlay : string, optional
            Change the default text overlay. Set to None to hide all text.
        """
        self.screenshotcountall = 0
        self.width, self.height  = size
        self.t, self.N = simulation.t, simulation.N
        self.orientation = orientation
        self.autorefresh = autorefresh
        self.orbits = orbits
        self.useroverlay = overlay
        self.simp = pointer(simulation)
        clibrebound.reb_display_copy_data.restype = c_int
        if scale is None:
            self.scale = simulation.display_data.contents.scale
        else:
            self.scale = scale
        self.count += 1

        super(Widget, self).__init__()

    def refresh(self, simp=None, isauto=0):
        """ 
        Manually refreshes a widget.
        
        Note that this function can also be called using the wrapper function of
        the Simulation object: sim.refreshWidgets(). 
        """

        if simp==None:
            simp = self.simp
        if self.autorefresh==0 and isauto==1:
            return
        sim = simp.contents
        size_changed = clibrebound.reb_display_copy_data(simp)
        clibrebound.reb_display_prepare_data(simp,c_int(self.orbits))
        if sim.N>0:
            self.particle_data = (c_char * (4*7*sim.N)).from_address(sim.display_data.contents.particle_data).raw
            if self.orbits:
                self.orbit_data = (c_char * (4*9*(sim.N-1))).from_address(sim.display_data.contents.orbit_data).raw
        if size_changed:
            #TODO: Implement better GPU size change
            pass
        if self.useroverlay==True:
            self.overlay = "REBOUND (%s), N=%d, t=%g"%(sim.integrator,sim.N,sim.t)
        elif self.useroverlay==None or self.useroverlay==False:
            self.overlay = ""
        else:
            self.overlay = self.useroverlay + ", N=%d, t=%g"%(sim.N,sim.t)
        self.N = sim.N
        self.t = sim.t
        self.count += 1

    def takeScreenshot(self, times=None, prefix="./screenshot", resetCounter=False, archive=None,mode="snapshot"):
        """
        Take one or more screenshots of the widget and save the images to a file. 
        The images can be used to create a video.

        This function cannot be called multiple times within one cell.

        Note: this is a new feature and might not work on all systems.
        It was tested on python 2.7.10 and 3.5.2 on MacOSX.
        
        Parameters
        ----------
        times : (float, list), optional
            If this argument is not given a screenshot of the widget will be made 
            as it is (without integrating the simulation). If a float is given, then the
            simulation will be integrated to that time and then a screenshot will 
            be taken. If a list of floats is given, the simulation will be integrated
            to each time specified in the array. A separate screenshot for 
            each time will be saved.
        prefix : (str), optional
            This string will be part of the output filename for each image.
            Follow by a five digit integer and the suffix .png. By default the
            prefix is './screenshot' which outputs images in the current
            directory with the filnames screenshot00000.png, screenshot00001.png...
            Note that the prefix can include a directory.
        resetCounter : (bool), optional
            Resets the output counter to 0. 
        archive : (rebound.SimulationArchive), optional
            Use a REBOUND SimulationArchive. Thus, instead of integratating the 
            Simulation from the current time, it will use the SimulationArchive
            to load a snapshot. See examples for usage.
        mode : (string), optional
            Mode to use when querying the SimulationArchive. See SimulationArchive
            documentation for details. By default the value is "snapshot".

        Examples
        --------

        First, create a simulation and widget. All of the following can go in 
        one cell.

        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> sim.add(m=1.e-3,x=1.,vy=1.)
        >>> w = sim.getWidget()
        >>> w

        The widget should show up. To take a screenshot, simply call 

        >>> w.takeScreenshot()

        A new file with the name screenshot00000.png will appear in the 
        current directory. 

        Note that the takeScreenshot command needs to be in a separate cell, 
        i.e. after you see the widget. 

        You can pass an array of times to the function. This allows you to 
        take multiple screenshots, for example to create a movie, 

        >>> times = [0,10,100]
        >>> w.takeScreenshot(times)

        """
        self.archive = archive
        if resetCounter:
            self.screenshotcountall = 0
        self.screenshotprefix = prefix
        self.screenshotcount = 0
        self.overlay = "REBOUND"
        self.screenshot = "" 
        if archive is None:
            if times is None:
                times = self.simp.contents.t
            try:
                # List
                len(times)
            except:
                # Float:
                times = [times]
            self.times = times
            self.observe(savescreenshot,names="screenshot")
            self.simp.contents.integrate(times[0])
            self.screenshotcount += 1 # triggers first screenshot
        else:
            if times is None:
                raise ValueError("Need times argument for archive mode.")
            try:
                len(times)
            except:
                raise ValueError("Need a list of times for archive mode.")
            self.times = times
            self.mode = mode
            self.observe(savescreenshot,names="screenshot")
            sim = archive.getSimulation(times[0],mode=mode)
            self.refresh(pointer(sim))
            self.screenshotcount += 1 # triggers first screenshot



            



    @staticmethod
    def getClientCode():
        return shader_code + js_code



import rebound
import unittest

class TestBoundary(unittest.TestCase):
    
    def test_open(self):
        sim = rebound.Simulation()
        sim.boundary = "open"
        sim.configure_box(10.)
        sim.add(m=0.1,x=1., vx=5.0)
        sim.add(m=0.1,x=-1., vx=-5.0)
        sim.add(m=0.1,y=1., vx=6.0)
        sim.add(m=0.1,x=-1., y=-1., vx=-3., vy=-3.)
        self.assertEqual(sim.N,4)
        sim.integrate(1.)
        self.assertEqual(sim.N,1)
        with self.assertRaises(rebound.NoParticles):
            sim.integrate(2.)
    
    def test_periodic(self):
        sim = rebound.Simulation()
        sim.boundary = "periodic"
        sim.configure_box(10.)
        sim.add(m=0.1,x=1., vx=5.0, vy=15.1, vz=26.)
        sim.integrate(1.)
        self.assertAlmostEqual(sim.particles[0].x,-4,delta=1e-16)
        sim.integrate(2.)
        self.assertAlmostEqual(sim.particles[0].x,1,delta=1e-16)
        self.assertEqual(sim.N,1)
    
    
if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import datetime
import socket
import warnings

class TestHorizons(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
    
    def tearDown(self):
        self.sim = None
    
    def test_earth(self):
        try:
            with warnings.catch_warnings(record=True) as w: 
                warnings.simplefilter("always")
                self.sim.add("Earth",date="2000-01-01 00:00")
                self.assertEqual(0,len(w))
            self.assertAlmostEqual(self.sim.particles[0].x,-0.17568959237103887,delta=1e-10)
            self.assertAlmostEqual(self.sim.particles[0].m,3.0404326480226416e-06,delta=1e-15)
        except socket.error: 
            print("Socket error. Most likely due to HORIZON being slow. Ignoring.")
            pass
    
    def test_notfound(self):
        with self.assertRaises(Exception):
            try:
                self.sim.add("BogusPlanet",date="2000-01-01 00:00")
            except socket.error: 
                print("Socket error. Most likely due to HORIZON being slow. Ignoring.")
                raise Exception("Socket error. Should have been bogus planet error. Ignoring")


if __name__ == "__main__":
    unittest.main()
